"""
Parse IgBLAST output and write out a tab-separated table.

IgBLAST must have been run with -outfmt "7 sseqid qstart qseq sstart sseq pident slen"

A few extra things are done in addition to parsing:
- The CDR3 is detected by using a regular expression
- The leader is detected within the sequence before the found V gene (by
  searching for the start codon).
- If the V sequence hit starts at base 2 in the reference, it is extended
  one to the left.
"""
import csv
import logging
from collections import namedtuple
import functools

from tinyalign import edit_distance, hamming_distance

from .utils import nt_to_aa
from .dna import reverse_complement


logger = logging.getLogger(__name__)


def none_if_na(s):
    """Return None if s == 'N/A'. Return s otherwise."""
    return None if s == 'N/A' else s


def gene_without_prefix(s):
    if s == "N/A":
        return None
    else:
        assert s.startswith("%")
        return s[1:]


def split_by_section(iterable, section_starts):
    """
    Parse a stream of lines into chunks of sections. When one of the lines
    starts with a string given in section_starts, a new section is started, and
    a tuple (head, lines) is returned where head is the matching line and lines
    contains a list of the lines following the section header, up to (but
    excluding) the next section header.

    Works a bit like str.split(), but on lines.
    """
    lines = None
    header = None
    for line in iterable:
        line = line.strip()
        for start in section_starts:
            if line.startswith(start):
                if header is not None:
                    yield (header, lines)
                header = line
                lines = []
                break
        else:
            if header is None:
                raise ParseError("Expected a line starting with one of {}".format(
                    ', '.join(section_starts)))
            lines.append(line)
    if header is not None:
        yield (header, lines)


JunctionVDJ = namedtuple('JunctionVDJ', 'v_end vd_junction d_region dj_junction j_start')
JunctionVJ = namedtuple('JunctionVJ', 'v_end vj_junction j_start')


class AlignmentSummary:
    """An alignment summary describes a framework region or complementarity-determining region
    (FR1/2/3, CDR1/2/3)"""

    def __init__(self, start, stop, length, matches, mismatches, gaps, percent_identity):
        self.start = start
        self.stop = stop
        self.length = length
        self.matches = matches
        self.mismatches = mismatches
        self.gaps = gaps
        if matches is not None:
            assert matches + mismatches + gaps == length
        if percent_identity is not None:
            assert matches is not None and length is not None
            assert abs(100. * matches / length - percent_identity) < 0.1
        self.percent_identity = percent_identity


class Hit:
    def __init__(
        self,
        subject_id: str,  # name of database record, such as "VH4.11"
        query_start: int,
        query_alignment: str,  # aligned part of the query, with '-' for deletions
        subject_start: int,
        subject_alignment: str,  # aligned part of reference, with '-' for insertions
        subject_length: int,  # total length of reference, depends only on subject_id
        percent_identity: float,
        evalue: float,
    ):
        assert len(subject_alignment) == len(query_alignment)
        self.subject_id = subject_id
        self.query_start = query_start
        self.query_alignment = query_alignment
        self.subject_start = subject_start
        self.subject_alignment = subject_alignment
        self.subject_length = subject_length
        self.percent_identity = percent_identity
        self.evalue = evalue

        # Derived attributes
        self.errors = self._errors(self.subject_alignment, self.query_alignment)
        self.query_sequence = self.query_alignment.replace('-', '')
        self.subject_sequence = self.subject_alignment.replace('-', '')

        assert abs(self.percent_identity - self._percent_identity()) < 0.01

    def extend_left_ungapped(self, query_sequence, subject_sequence):
        """
        Extend this hit to the left until it reaches the first nucleotide of the subject sequence.
        Used for extending V hits to the 5' end.
        """
        query_bases = []
        subject_bases = []

        while self.subject_start > 0 and self.query_start > 0:
            self.query_start -= 1
            self.subject_start -= 1
            query_base = query_sequence[self.query_start]
            query_bases.append(query_base)
            if subject_sequence is not None:
                subject_base = subject_sequence[self.subject_start]
            else:
                subject_base = 'N'
            subject_bases.append(subject_base)

        query_bases = ''.join(query_bases[::-1])
        subject_bases = ''.join(subject_bases[::-1])
        self.query_alignment = query_bases + self.query_alignment
        self.subject_alignment = subject_bases + self.subject_alignment
        self.query_sequence = query_bases + self.query_sequence
        self.subject_sequence = subject_bases + self.subject_sequence
        self.errors += self._errors(query_bases, subject_bases)
        self.percent_identity = self._percent_identity()

    def _percent_identity(self):
        """This is how IgBLAST computes percent identity"""
        matches = len(self.subject_alignment) - self.errors
        return 100. * matches / len(self.subject_alignment)

    def covered(self):
        """
        Return fraction of bases in the original subject sequence that are
        covered by this hit.
        """
        return len(self.subject_sequence) / self.subject_length

    @property
    def query_end(self):
        return self.query_start + len(self.query_sequence)

    @property
    def subject_end(self):
        return self.subject_start + len(self.subject_sequence)

    @staticmethod
    def _errors(alignment1, alignment2):
        return sum(a != b for a, b in zip(alignment1, alignment2))

    def query_position(self, reference_position):
        """
        Given a position on the reference, return the same position but relative to
        the full query sequence.
        """
        # Iterate over alignment columns
        ref_pos = self.subject_start
        query_pos = self.query_start
        if ref_pos == reference_position:
            return query_pos
        for ref_c, query_c in zip(self.subject_alignment, self.query_alignment):
            if ref_c != '-':
                ref_pos += 1
            if query_c != '-':
                query_pos += 1
            if ref_pos == reference_position:
                return query_pos
        return None


def parse_header(header):
    """
    Extract size= and barcode= fields from the FASTA/FASTQ header line

    >>> parse_header("name;size=12;barcode=ACG;")
    ('name', 12, 'ACG')
    >>> parse_header("another name;size=200;foo=bar;")
    ('another name', 200, None)
    """
    fields = header.split(';')
    query_name = fields[0]
    size = barcode = None
    for field in fields[1:]:
        if field == '':
            continue
        if '=' in field:
            key, value = field.split('=', maxsplit=1)
            if key == 'size':
                size = int(value)
            elif key == 'barcode':
                barcode = value
    return query_name, size, barcode


class IgBlastRecord:
    def __init__(
        self,
        full_sequence,
        query_name,
        alignments,
        hits,
        v_gene,
        d_gene,
        j_gene,
        chain,
        has_stop,
        in_frame,
        is_productive,
        strand,
        junction
    ):
        self.full_sequence = full_sequence
        self.query_name = query_name
        self.alignments = alignments
        self.hits = hits
        self.v_gene = v_gene
        self.d_gene = d_gene
        self.j_gene = j_gene
        self.chain = chain
        self.has_stop = has_stop
        self.in_frame = in_frame
        self.is_productive = is_productive
        self.strand = strand
        self.junction = junction

    def region_sequence(self, region):
        """
        Return the nucleotide sequence of a named region. Allowed names are:
        CDR1, CDR2, CDR3, FR1, FR2, FR3. Sequences are extracted from the full read
        using begin and end coordinates from IgBLAST’s "alignment summary" table.
        """
        if region not in ("CDR1", "CDR2", "CDR3", "FR1", "FR2", "FR3"):
            raise KeyError(f"Region {region!r} not allowed")
        alignment = self.alignments.get(region, None)
        if alignment is None:
            return None
        if alignment.start is None or alignment.stop is None:
            return None
        return self.full_sequence[alignment.start:alignment.stop]

    def __repr__(self):
        return 'IgBlastRecord(query_name={query_name!r}, ' \
            'v_gene={v_gene!r}, d_gene={d_gene!r}, j_gene={j_gene!r}, chain={chain!r}, ...)'.format(
                **vars(self))


class Region:
    """A CDR or FR region in a V(D)J rearranged sequence (FR1, CDR1, FR2, CDR2, FR3, CDR3, FR4)"""

    def __init__(self, nt_sequence, nt_reference, aa_reference=None, percent_identity=None):
        self.nt_sequence = nt_sequence
        self.aa_sequence = nt_to_aa(nt_sequence) if nt_sequence else None
        self.nt_reference = nt_reference
        if aa_reference is None and nt_reference is not None:
            aa_reference = nt_to_aa(nt_reference)
        self.aa_reference = aa_reference
        self.aa_mutations = self._aa_mutations()
        self.percent_identity = percent_identity
        if percent_identity is None:
            self.percent_identity = self._percent_identity()

    def _percent_identity(self):
        # FIXME This is not quite how IgBLAST computes percent identity
        if not self.nt_reference or not self.nt_sequence:
            return None
        dist = edit_distance(self.nt_reference, self.nt_sequence)
        return 100. - 100. * dist / len(self.nt_reference)

    def _aa_mutations(self):
        # Earlier versions of this code used edit distance to compute the number of mutations,
        # but some FR1 alignments are reported with a frameshift by IgBLAST. By requiring that
        # reference and query lengths are identical, we can filter out these cases (and use
        # Hamming distance to get some speedup)
        if (
            not self.aa_reference
            or not self.aa_sequence
            or len(self.nt_sequence) != len(self.nt_reference)
        ):
            return None
        dist = hamming_distance(self.aa_reference, self.aa_sequence)

        # If the mutation rate is still obviously too high, assume something went
        # wrong and ignore the computed value
        if dist / len(self.aa_reference) >= 0.8:
            return None
        return dist

    def aa_mutation_rate(self):
        if self.aa_mutations is None or not self.aa_reference:
            return None
        return 100. * self.aa_mutations / len(self.aa_reference)

    def nt_mutation_rate(self):
        """Return nucleotide-level mutation rate in percent"""
        if self.percent_identity is None:
            return None
        return 100. - self.percent_identity


class ExtendedIgBlastRecord(IgBlastRecord):
    """
    This extended record does a few extra things:
    - The CDR3 is detected
    - The leader is detected within the sequence before the found V gene (by
    searching for the start codon).
    - If the V sequence hit starts not at base 1 in the reference, it is extended
    to the left.
    """
    # TODO move computation of cdr3_sequence, vdj_sequence into constructor
    # TODO maybe make all coordinates relative to full sequence

    # Order of columns (use with asdict())
    columns = [
        'count',
        'V_gene',
        'D_gene',
        'J_gene',
        'chain',
        'stop',
        'V_covered',
        'D_covered',
        'J_covered',
        'V_evalue',
        'D_evalue',
        'J_evalue',
        'FR1_SHM',
        'CDR1_SHM',
        'FR2_SHM',
        'CDR2_SHM',
        'FR3_SHM',
        'FR4_SHM',
        'V_SHM',
        'J_SHM',
        'V_aa_mut',
        'J_aa_mut',
        'FR1_aa_mut',
        'CDR1_aa_mut',
        'FR2_aa_mut',
        'CDR2_aa_mut',
        'FR3_aa_mut',
        'V_errors',
        'D_errors',
        'J_errors',
        'UTR',
        'leader',
        'CDR1_nt',
        'CDR1_aa',
        'CDR2_nt',
        'CDR2_aa',
        'CDR3_nt',
        'CDR3_aa',
        'V_nt',
        'V_aa',
        'V_end',
        'V_CDR3_start',
        'VD_junction',
        'D_region',
        'DJ_junction',
        'J_nt',
        'VDJ_nt',
        'VDJ_aa',
        'name',
        'barcode',
        'genomic_sequence',
    ]

    CHAINS = {
        'VH': 'heavy', 'VK': 'kappa', 'VL': 'lambda',
        'VA': 'alpha', 'VB': 'beta',
        'VG': 'gamma', 'VD': 'delta'
    }

    def __init__(self, database, **kwargs):
        super().__init__(**kwargs)
        self.query_name, self.size, self.barcode = parse_header(self.query_name)
        self.genomic_sequence = self.full_sequence
        self._database = database
        if 'V' in self.hits:
            subject_sequence = self._database.v[self.hits['V'].subject_id]
            self.hits['V'].extend_left_ungapped(self.full_sequence, subject_sequence)
        self.utr, self.leader = self._utr_leader()
        self.alignments['CDR3'] = self._find_cdr3()

        self.regions = {
            name: self._make_region(name) for name in
            ('FR1', 'FR2', 'FR3', 'CDR1', 'CDR2', 'CDR3')}
        self.regions['FR4'] = self._make_fr4_region()
        self.vdj_sequence = self._make_vdj_sequence()

    def _make_region(self, name: str):
        nt_sequence = self.region_sequence(name)
        if self.v_gene in self._database.v_regions_nt:
            nt_reference = self._database.v_regions_nt[self.v_gene].get(name)
        else:
            nt_reference = None
        if self.v_gene in self._database.v_regions_aa:
            aa_reference = self._database.v_regions_aa[self.v_gene].get(name)
        else:
            aa_reference = None
        if self.alignments.get(name, None) is not None:
            percent_identity = self.alignments[name].percent_identity
        else:
            percent_identity = None
        return Region(nt_sequence, nt_reference, aa_reference, percent_identity)

    def _make_fr4_region(self):
        if 'J' not in self.hits:
            return Region(None, None)
        j_subject_id = self.hits['J'].subject_id
        if self.chain not in self.CHAINS:
            return Region(None, None)
        cdr3_ref_end = self._database.j_cdr3_end(j_subject_id, self.CHAINS[self.chain])
        if cdr3_ref_end is None:
            return Region(None, None)
        cdr3_query_end = self.hits['J'].query_position(reference_position=cdr3_ref_end)
        if cdr3_query_end is None:
            return Region(None, None)

        query = self.full_sequence[cdr3_query_end:self.hits['J'].query_end]
        ref = self._database.j[j_subject_id][cdr3_ref_end:self.hits['J'].subject_end]

        return Region(query, ref)

    def _make_vdj_sequence(self):
        if 'V' not in self.hits or 'J' not in self.hits:
            return None
        hit_v = self.hits['V']
        hit_j = self.hits['J']
        vdj_start = hit_v.query_start
        vdj_stop = hit_j.query_start + len(hit_j.query_sequence)
        return self.full_sequence[vdj_start:vdj_stop]

    @property
    def v_cdr3_start(self):
        """Start of CDR3 within V"""
        if 'V' not in self.hits or self.alignments['CDR3'] is None:
            return 0
        v_start = self.hits['V'].query_start
        cdr3_start = self.alignments['CDR3'].start
        return cdr3_start - v_start

    def _utr_leader(self):
        """
        Split the sequence before the V gene match into UTR and leader by
        searching for the start codon.
        """
        if 'V' not in self.hits:
            return None, None
        before_v = self.full_sequence[:self.hits['V'].query_start]

        # Search for the start codon
        for offset in (0, 1, 2):
            for i in range(66, 42, -3):
                if before_v[-i + offset:-i + 3 + offset] == 'ATG':
                    return before_v[:-i + offset], before_v[-i + offset:]
        return None, None

    def _find_cdr3(self):
        """
        Return a repaired AlignmentSummary object that describes the CDR3 region.
        Return None if no CDR3 detected.
        """
        if 'V' not in self.hits or 'J' not in self.hits:
            return None
        if self.chain not in self.CHAINS:
            return None

        # CDR3 start
        cdr3_ref_start = self._database.v_cdr3_start(self.hits['V'].subject_id, self.CHAINS[self.chain])
        if cdr3_ref_start is None:
            return None
        cdr3_query_start = self.hits['V'].query_position(reference_position=cdr3_ref_start)
        if cdr3_query_start is None:
            # Alignment is not long enough to cover CDR3 start position; try to rescue it
            # by assuming that the alignment would continue without indels.
            hit = self.hits['V']
            cdr3_query_start = hit.query_end + (cdr3_ref_start - hit.subject_end)

        # CDR3 end
        cdr3_ref_end = self._database.j_cdr3_end(self.hits['J'].subject_id, self.CHAINS[self.chain])
        if cdr3_ref_end is None:
            return None

        cdr3_query_end = self.hits['J'].query_position(reference_position=cdr3_ref_end)
        if cdr3_query_end is None:
            return None

        return AlignmentSummary(start=cdr3_query_start, stop=cdr3_query_end, length=None,
            matches=None, mismatches=None, gaps=None, percent_identity=None)

    def fr4_aa_mutation_rate(self):
        if 'J' not in self.hits:
            return None
        j_subject_id = self.hits['J'].subject_id
        if self.chain not in self.CHAINS:
            return None
        cdr3_ref_end = self._database.j_cdr3_end(j_subject_id, self.CHAINS[self.chain])
        if cdr3_ref_end is None:
            return None
        cdr3_query_end = self.hits['J'].query_position(reference_position=cdr3_ref_end)
        if cdr3_query_end is None:
            return None

        query = self.full_sequence[cdr3_query_end:self.hits['J'].query_end]
        try:
            query_aa = nt_to_aa(query)
        except ValueError:
            return None
        ref = self._database.j[j_subject_id][cdr3_ref_end:self.hits['J'].subject_end]
        try:
            ref_aa = nt_to_aa(ref)
        except ValueError:
            return None
        if not ref_aa:
            return None
        return 100. * edit_distance(ref_aa, query_aa) / len(ref_aa)

    def v_aa_mutation_rate(self):
        """
        TODO This returns actually the total mutation rate of the FR1, CDR1, FR2, CDR2, FR3 regions
        (The CDR3 part of V is excluded.)
        """
        mutations = 0
        length = 0
        for name in ('FR1', 'CDR1', 'FR2', 'CDR2', 'FR3'):
            region = self.regions.get(name)
            if region is None:
                return None
            if region.aa_reference is None or region.aa_mutations is None:
                return None
            mutations += region.aa_mutations
            length += len(region.aa_reference)

        return 100. * mutations / length

    def asdict(self):
        """
        Return a flattened representation of this record as a dictionary.
        The dictionary can then be used with e.g. a csv.DictWriter or
        pandas.DataFrame.from_items.
        """
        if 'V' in self.hits:
            v_nt = self.hits['V'].query_sequence
            v_aa = nt_to_aa(v_nt)
            v_shm = 100. - self.hits['V'].percent_identity
            v_errors = self.hits['V'].errors
            v_covered = 100. * self.hits['V'].covered()
            v_evalue = self.hits['V'].evalue
        else:
            v_nt = None
            v_aa = None
            v_shm = None
            v_errors = None
            v_covered = None
            v_evalue = None
        if 'D' in self.hits:
            d_errors = self.hits['D'].errors
            d_covered = 100. * self.hits['D'].covered()
            d_evalue = self.hits['D'].evalue
        else:
            d_errors = None
            d_covered = None
            d_evalue = None
        if 'J' in self.hits:
            j_nt = self.hits['J'].query_sequence
            j_shm = 100. - self.hits['J'].percent_identity
            j_errors = self.hits['J'].errors
            j_covered = 100. * self.hits['J'].covered()
            j_evalue = self.hits['J'].evalue
        else:
            j_nt = None
            j_shm = None
            j_errors = None
            j_covered = None
            j_evalue = None
        v_end = getattr(self.junction, 'v_end', None)
        vd_junction = getattr(self.junction, 'vd_junction', None)
        d_region = getattr(self.junction, 'd_region', None)
        dj_junction = getattr(self.junction, 'dj_junction', None)

        return dict(
            count=self.size,
            V_gene=self.v_gene,
            D_gene=self.d_gene,
            J_gene=self.j_gene,
            chain=self.chain,
            stop=self.has_stop,
            V_covered=v_covered,
            D_covered=d_covered,
            J_covered=j_covered,
            V_evalue=v_evalue,
            D_evalue=d_evalue,
            J_evalue=j_evalue,
            FR1_SHM=self.regions['FR1'].nt_mutation_rate(),
            CDR1_SHM=self.regions['CDR1'].nt_mutation_rate(),
            FR2_SHM=self.regions['FR2'].nt_mutation_rate(),
            CDR2_SHM=self.regions['CDR2'].nt_mutation_rate(),
            FR3_SHM=self.regions['FR3'].nt_mutation_rate(),
            FR4_SHM=self.regions['FR4'].nt_mutation_rate(),
            V_SHM=v_shm,
            J_SHM=j_shm,
            V_aa_mut=self.v_aa_mutation_rate(),
            J_aa_mut=self.regions['FR4'].aa_mutation_rate(),  # TODO J vs FR4
            FR1_aa_mut=self.regions['FR1'].aa_mutation_rate(),
            CDR1_aa_mut=self.regions['CDR1'].aa_mutation_rate(),
            FR2_aa_mut=self.regions['FR2'].aa_mutation_rate(),
            CDR2_aa_mut=self.regions['CDR2'].aa_mutation_rate(),
            FR3_aa_mut=self.regions['FR3'].aa_mutation_rate(),
            # FR4_aa_mut=aa_rates['FR4'],  # TODO
            V_errors=v_errors,
            D_errors=d_errors,
            J_errors=j_errors,
            UTR=self.utr,
            leader=self.leader,
            CDR1_nt=self.regions['CDR1'].nt_sequence,
            CDR1_aa=self.regions['CDR1'].aa_sequence,
            CDR2_nt=self.regions['CDR2'].nt_sequence,
            CDR2_aa=self.regions['CDR2'].aa_sequence,
            CDR3_nt=self.regions['CDR3'].nt_sequence,
            CDR3_aa=self.regions['CDR3'].aa_sequence,
            V_nt=v_nt,
            V_aa=v_aa,
            V_end=v_end,
            V_CDR3_start=self.v_cdr3_start,
            VD_junction=vd_junction,
            D_region=d_region,
            DJ_junction=dj_junction,
            J_nt=j_nt,
            VDJ_nt=self.vdj_sequence,
            VDJ_aa=nt_to_aa(self.vdj_sequence) if self.vdj_sequence else None,
            name=self.query_name,
            barcode=self.barcode,
            genomic_sequence=self.genomic_sequence,
        )


class ParseError(Exception):
    pass


class IgBlastParser:
    """
    Parser for IgBLAST results. Works only when IgBLAST was run with
    the option -outfmt "7 sseqid qstart qseq sstart sseq pident slen".
    """
    BOOL = {'Yes': True, 'No': False, 'N/A': None}
    FRAME = {'In-frame': True, 'Out-of-frame': False, 'N/A': None}
    SECTIONS = frozenset([
        '# Query:',
        '# V-(D)-J rearrangement summary',
        '# V-(D)-J junction details',
        '# Alignment summary',
        '# Hit table',
        'Total queries = ',
    ])

    def __init__(self, sequences, igblast_lines, database=None):
        """
        If a database is given, iterating over this object will
        yield ExtendedIgBlastRecord objects, otherwise 'normal' IgBlastRecord objects
        """
        self._sequences = sequences
        self._igblast_lines = igblast_lines
        self._database = database
        if self._database is None:
            self._create_record = IgBlastRecord
        else:
            self._create_record = functools.partial(ExtendedIgBlastRecord, database=self._database)

    def __iter__(self):
        """
        Yield (Extended-)IgBlastRecord objects
        """
        zipped = zip(self._sequences, split_by_section(self._igblast_lines, ['# IGBLASTN']))
        for fasta_record, (record_header, record_lines) in zipped:
            # 'IGBLASTN 2.5.1+': IgBLAST 1.6.1
            assert record_header in {
                '# IGBLASTN 2.2.29+',  # IgBLAST 1.4.0
                '# IGBLASTN 2.3.1+',  # IgBLAST 1.5.0
                '# IGBLASTN 2.6.1+',  # IgBLAST 1.7.0
                '# IGBLASTN',  # IgBLAST 1.10
            }
            yield self._parse_record(record_lines, fasta_record)

    def _parse_record(self, record_lines, fasta_record):
        """
        Parse a single IgBLAST record
        """
        hits = dict()
        # All of the sections are optional, so we need to set default values here.
        query_name = None
        junction = None
        v_gene, d_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = [None] * 8
        alignments = dict()
        for section, lines in split_by_section(record_lines, self.SECTIONS):
            if section.startswith('# Query: '):
                query_name = section.split(': ')[1]
            elif section.startswith('# V-(D)-J rearrangement summary'):
                fields = lines[0].split('\t')
                if len(fields) == 7:
                    # No D assignment
                    v_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = fields
                    d_gene = None
                else:
                    v_gene, d_gene, j_gene, chain, has_stop, in_frame, is_productive, strand = fields
                    d_gene = gene_without_prefix(d_gene)
                v_gene = gene_without_prefix(v_gene)
                j_gene = gene_without_prefix(j_gene)
                chain = none_if_na(chain)
                has_stop = self.BOOL[has_stop]
                in_frame = self.FRAME[in_frame]
                is_productive = self.BOOL[is_productive]
                strand = strand if strand in '+-' else None
            elif section.startswith('# V-(D)-J junction details'):
                fields = lines[0].split('\t')
                if len(fields) == 5:
                    junction = JunctionVDJ(
                        v_end=fields[0],
                        vd_junction=fields[1],
                        d_region=fields[2],
                        dj_junction=fields[3],
                        j_start=fields[4]
                    )
                else:
                    junction = JunctionVJ(
                        v_end=fields[0],
                        vj_junction=fields[1],
                        j_start=fields[2])
            elif section.startswith('# Alignment summary'):
                for line in lines:
                    fields = line.split('\t')
                    if len(fields) == 8 and fields[0] != 'Total':
                        summary = self._parse_alignment_summary(fields[1:])
                        region_name, _, imgt = fields[0].partition('-')
                        assert imgt in ('IMGT', 'IMGT (germline)')
                        alignments[region_name] = summary
            elif section.startswith('# Hit table'):
                for line in lines:
                    if not line or line.startswith('#'):
                        continue
                    hit, gene = self._parse_hit(line)
                    assert gene in ('V', 'D', 'J')
                    assert gene not in hits, "Two hits for same gene found"
                    hits[gene] = hit
            elif section.startswith('Total queries = '):
                continue

        assert fasta_record.name == query_name
        full_sequence = fasta_record.sequence.upper()
        if strand == '-':
            full_sequence = reverse_complement(full_sequence)

        if __debug__:
            for gene in ('V', 'D', 'J'):
                if gene not in hits:
                    continue
                hit = hits[gene]
                qsequence = hit.query_sequence

                # IgBLAST removes the trailing semicolon (why, oh why??)
                qname = query_name[:-1] if query_name.endswith(';') else query_name
                assert chain in (None, 'VL', 'VH', 'VK', 'NON', 'VA', 'VB', 'VG', 'VD'), chain
                assert qsequence == full_sequence[hit.query_start:hit.query_start+len(qsequence)]

        return self._create_record(
            query_name=query_name,
            alignments=alignments,
            v_gene=v_gene,
            d_gene=d_gene,
            j_gene=j_gene,
            chain=chain,
            has_stop=has_stop,
            in_frame=in_frame,
            is_productive=is_productive,
            strand=strand,
            hits=hits,
            full_sequence=full_sequence,
            junction=junction)

    def _parse_alignment_summary(self, fields):
        start, stop, length, matches, mismatches, gaps = (int(v) for v in fields[:6])
        percent_identity = float(fields[6])
        assert abs(percent_identity - 100. * matches / length) < 0.1
        # Note length is not necessarily equal to stop - start. Not sure why.
        return AlignmentSummary(
            start=start - 1,
            stop=stop,
            length=length,
            matches=matches,
            mismatches=mismatches,
            gaps=gaps,
            percent_identity=percent_identity
        )

    def _parse_hit(self, line):
        """
        Parse a line of the "Hit table" section and return a tuple (hit, gene)
        where hit is a Hit object.
        """
        (gene, subject_id, query_start, query_alignment, subject_start, subject_alignment,
            percent_identity, subject_length, evalue) = line.split('\t')
        # Names have been mangled by adding a '%' to the beginning. Otherwise, IgBLAST
        # may recognize sequence ids such as AB123456 as being an accession and mangle
        # them to conform to BLAST’s naming scheme.
        # Undo this here.
        assert subject_id.startswith('%')
        subject_id = subject_id[1:]
        query_start = int(query_start) - 1
        subject_start = int(subject_start) - 1
        subject_length = int(subject_length)  # Length of original subject sequence
        percent_identity = float(percent_identity)
        evalue = float(evalue)
        hit = Hit(subject_id, query_start, query_alignment, subject_start,
            subject_alignment, subject_length, percent_identity, evalue)
        return hit, gene


class TableWriter:
    def __init__(self, file):
        self._file = file
        self._writer = csv.DictWriter(file, fieldnames=ExtendedIgBlastRecord.columns, delimiter='\t')
        self._writer.writeheader()

    @staticmethod
    def yesno(v):
        """
        Return "yes", "no" or None for boolean value v, which may also be None.
        """
        if v is None:
            return None
        return ["no", "yes"][v]

    def write(self, d):
        """
        Write the IgBLAST record (must be given as dictionary) to the output
        file.
        """
        d = d.copy()
        d['stop'] = self.yesno(d['stop'])
        for name in ('V_covered', 'D_covered', 'J_covered',
                'FR1_SHM', 'CDR1_SHM', 'FR2_SHM', 'CDR2_SHM', 'FR3_SHM', 'FR4_SHM',
                'V_SHM', 'J_SHM', 'V_aa_mut', 'J_aa_mut',
                'FR1_aa_mut', 'CDR1_aa_mut', 'FR2_aa_mut', 'CDR2_aa_mut', 'FR3_aa_mut'):
            if d[name] is not None:
                d[name] = '{:.1f}'.format(d[name])
        for name in ('V_evalue', 'D_evalue', 'J_evalue'):
            if d[name] is not None:
                d[name] = '{:G}'.format(d[name])
        self._writer.writerow(d)



"""
Filter V gene candidates (germline and pre-germline filter)

After candidates for novel V genes have been found with the 'discover'
subcommand, this script is used to filter the candidates and make sure that
only true germline genes remain ("germline filter" and "pre-germline filter").
The following filtering and processing steps are performed on each candidate
separately:

* Discard sequences with N bases
* Discard sequences that come from a consensus over too few source sequences (unless whitelisted)
* Discard sequences with too few unique CDR3s (CDR3_clusters column)
* Discard sequences with too few unique Js (Js_exact column)
* Discard sequences identical to one of the database sequences (if DB given)
* Discard sequences that contain a stop codon (has_stop column) (unless whitelisted)

The following criteria involve comparison of candidates against each other:

* Discard sequences that are identical (unless whitelisted)
* Discard sequences that are cross-mapping artifacts
* Discard sequences that have a too low allele ratio
* Discard sequences with too few unique Ds relative to other alleles (Ds_exact column)

If you provide a whitelist of sequences, then the candidates that appear on it
* are not checked for the cluster size criterion,
* are never considered near-duplicates,
* are allowed to contain a stop codon.

The filtered table is written to standard output.
"""
import sys
import logging
from typing import NamedTuple

import dnaio
import pandas as pd
from tinyalign import edit_distance

from ..utils import UniqueNamer, is_same_gene, ChimeraFinder

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--cluster-size', type=int, metavar='N', default=0,
        help='Consensus must represent at least N sequences. '
        'Default: %(default)s')
    arg('--cross-mapping-ratio', type=float, metavar='RATIO', default=0.02,
        help='Ratio for detection of cross-mapping artifacts. Default: %(default)s')
    arg('--clonotype-ratio', '--allele-ratio', type=float, metavar='RATIO', default=0.1,
        help='Required ratio of "clonotypes" counts between alleles. '
            'Works only for genes named "NAME*ALLELE". Default: %(default)s')
    arg('--exact-ratio', type=float, metavar='RATIO', default=0.1,
        help='Required ratio of "exact" counts between alleles. '
            'Works only for genes named "NAME*ALLELE". Default: %(default)s')
    arg('--cdr3-shared-ratio', type=float, metavar='RATIO', default=1.0,
        help='Maximum allowed CDR3_shared_ratio. Default: %(default)s')
    arg('--minimum-db-diff', '-b', type=int, metavar='N', default=0,
        help='Sequences must have at least N differences to the database '
        'sequence. Default: %(default)s')
    arg('--maximum-N', '-N', type=int, metavar='COUNT', default=0,
        help='Sequences must have at most COUNT "N" bases. Default: %(default)s')
    arg('--unique-cdr3s', '--unique-CDR3', '--cdr3s', type=int, metavar='N', default=1,
        help='Sequences must have at least N unique CDR3s within exact sequence matches. '
        'Default: %(default)s')
    # The default for unique-J is 0 because we might work on data without
    # any assigned J genes.
    arg('--unique-J', type=int, metavar='N', default=0,
        help='Sequences must have at least N unique Js within exact sequence matches. '
        'Default: %(default)s')
    arg('--unique-D-ratio', type=float, metavar='RATIO', default=None,
        help='Discard a sequence if another allele of this gene exists '
        'such that the ratio between their Ds_exact is less than RATIO')
    arg('--unique-D-threshold', type=int, metavar='THRESHOLD', default=10,
        help='Apply the --unique-D-ratio filter only if the Ds_exact of the other '
        'allele is at least THRESHOLD')
    arg('--allow-stop', action='store_true', default=False,
        help='Allow stop codons in sequences (uses the has_stop column).'
            'Default: Do not allow stop codons.')
    # arg('--allow-chimeras', action='store_true', default=False,
    #     help='Do not filter out chimeric sequences. Default: Discard chimeras')
    arg('--whitelist', metavar='FASTA', default=[], action='append',
        help='Sequences that are never discarded or merged with others, '
            'even if criteria for discarding them would apply (except cross-mapping artifact '
            'removal, which is always performed).')
    arg('--fasta', metavar='FILE', help='Write new database in FASTA format to FILE')
    arg('--annotate', metavar='FILE',
        help='Write candidates.tab with filter annotations to FILE')
    arg('tables', metavar='CANDIDATES.TAB',
        help='Tables (one or more) created by the "discover" command',
        nargs='+')


class Candidate(NamedTuple):
    sequence: str
    name: str
    clonotypes: int
    exact: int
    Ds_exact: int
    cluster_size: int
    whitelisted: bool
    is_database: bool
    cluster_size_is_accurate: bool
    cdr3_start: int
    index: int


class IdenticalSequenceFilter:
    """
    Filter a candidate if it is identical to another one or if it is a truncated version
    of another one
    """
    @staticmethod
    def should_discard(ref: Candidate, candidate: Candidate, _same_gene: bool):
        if candidate.whitelisted:
            return False
        if ref.cluster_size < candidate.cluster_size:
            return False
        if ref.sequence == candidate.sequence:
            return f'identical_to={ref.name}'
        if ref.sequence.startswith(candidate.sequence):
            return f'identical_to={ref.name},truncated'

        return False


class CrossMappingFilter:
    """
    Filter a candidate if it is a cross-mapping artifact
    """
    def __init__(self, cross_mapping_ratio: float):
        """Check for cross-mapping"""
        self._ratio = cross_mapping_ratio

    def should_discard(self, reference: Candidate, candidate: Candidate, _same_gene: bool):
        """
        Compare a candidate to a reference candidate and decide whether it should be discarded.

        :param reference: The reference candidate. The decision this function makes is not about that one.
        :param candidate: The candidate on which to decide.
        :param dist: Edit distance between candidates
        :return: False if the candidate should be kept. Otherwise, the candidate shoud be discarded
        and a non-empty string with a reason describing why is returned.
        """
        # When computing edit distance between the two sequences, ignore the
        # bases in the 3' end that correspond to the CDR3
        s_no_cdr3 = reference.sequence[:reference.cdr3_start]
        t_no_cdr3 = candidate.sequence[:candidate.cdr3_start]
        if len(s_no_cdr3) != len(t_no_cdr3):
            t_prefix = t_no_cdr3[:len(s_no_cdr3)]
            t_suffix = t_no_cdr3[-len(s_no_cdr3):]
            dist_prefix = edit_distance(s_no_cdr3, t_prefix, 1)
            dist_suffix = edit_distance(s_no_cdr3, t_suffix, 1)  # TODO prefix and suffix?
            dist_no_cdr3 = min(dist_prefix, dist_suffix)
        else:
            dist_no_cdr3 = edit_distance(s_no_cdr3, t_no_cdr3, 1)

        if dist_no_cdr3 > 1:
            # Cross-mapping is unlikely if the edit distance is larger than 1
            return None
        if not reference.is_database or not candidate.is_database:
            # Cross-mapping can only occur if both sequences are in the database
            return None

        total_count = (reference.cluster_size + candidate.cluster_size)
        if total_count == 0:
            return False
        ratio = candidate.cluster_size / total_count
        if candidate.cluster_size_is_accurate and ratio < self._ratio:
            # candidate is probably a cross-mapping artifact of the higher-expressed ref
            return f'xmap_ratio={ratio:.4f},other={reference.name}'
        return False


class ClonotypeAlleleRatioFilter:
    """
    Clonotype allele ratio filter. Somewhat similar to cross-mapping, but
    this uses sequence names to decide whether two genes can be
    alleles of each other and the ratio is between the CDR3_clusters values
    """
    def __init__(self, clonotype_ratio: float):
        self._ratio = clonotype_ratio

    def should_discard(self, ref: Candidate, candidate: Candidate, same_gene: bool):
        if not same_gene:
            return False
        if ref.clonotypes == 0:
            return False
        ratio = candidate.clonotypes / ref.clonotypes
        if ratio < self._ratio:
            # Clonotype allele ratio too low
            return f'clonotype_ratio={ratio:.4f},other={ref.name}'
        return False


class ExactRatioFilter:
    """Exact V sequence occurrence allele ratio"""

    def __init__(self, exact_ratio: float):
        self._ratio = exact_ratio

    def should_discard(self, ref: Candidate, candidate: Candidate, same_gene: bool):
        if not same_gene:
            return False
        if ref.exact == 0:
            return False
        ratio = candidate.exact / ref.exact
        if ratio < self._ratio:
            # Allele ratio of exact occurrences too low
            return f'ex_occ_ratio={ratio:.1f},other={ref.name}'
        return False


class UniqueDRatioFilter:
    def __init__(self, unique_d_ratio: float, unique_d_threshold: int):
        self._unique_d_ratio = unique_d_ratio
        self._unique_d_threshold = unique_d_threshold

    def should_discard(self, ref: Candidate, candidate: Candidate, same_gene: bool):
        if not same_gene:
            return False
        if ref.cluster_size < candidate.cluster_size or ref.Ds_exact < self._unique_d_threshold:
            # TODO comment
            return False
        ratio = candidate.Ds_exact / ref.Ds_exact
        if ratio < self._unique_d_ratio:
            # Ds_exact ratio too low
            return f'Ds_exact_ratio={ratio:.1f},other={ref.name}'
        return False


class CandidateFilterer:
    """Apply filters that compare candidates to each other to a list of candidates"""

    def __init__(self, filters):
        self._filters = filters

    def should_discard(self, reference: Candidate, candidate: Candidate):
        """
        Given a candidate to keep (reference), decide whether another one should be discarded
        """
        same_gene = is_same_gene(reference.name, candidate.name)
        for filter_ in self._filters:
            reason = filter_.should_discard(reference, candidate, same_gene)
            if reason:
                return reason

        # None of the filters decided to discard the candidate, so keep it
        return False

    @staticmethod
    def cluster_size_is_accurate(row):
        return bool(set(row.cluster.split(';')) & {'all', 'db'})

    @classmethod
    def row_to_candidate(cls, row):
        return Candidate(
            sequence=row['consensus'],
            name=row['name'],
            clonotypes=row['clonotypes'],
            exact=row['exact'],
            Ds_exact=row['Ds_exact'],
            cluster_size=row['cluster_size'],
            whitelisted=row['whitelist_diff'] == 0,
            is_database=row['database_diff'] == 0,
            cluster_size_is_accurate=cls.cluster_size_is_accurate(row),
            cdr3_start=row.get('CDR3_start', 10000),  # TODO backwards compatibility
            index=row.name,
        )

    def apply(self, table):
        """
        Run the filters and update the why_filtered and is_filtered columns appropriately for
        the filtered candidates in the given table (in place). No rows are removed.
        """
        candidates = [self.row_to_candidate(row) for _, row in table.iterrows()]
        for reference in candidates:
            if table.loc[reference.index, 'is_filtered'] > 0:
                continue
            for candidate in candidates:
                if reference is candidate:
                    # Don’t compare a candidate to itself
                    continue
                reason = self.should_discard(reference, candidate)
                if reason:
                    table.loc[candidate.index, 'why_filtered'] += reason + ';'
                    table.loc[candidate.index, 'is_filtered'] += 1


class Whitelist:
    def __init__(self):
        self._sequences = dict()

    def add_fasta(self, path):
        with dnaio.open(path) as fr:
            for record in fr:
                self._sequences[record.sequence.upper()] = record.name

    def closest(self, sequence):
        """
        Search for the whitelist sequence that is closest to the given sequence.

        Return tuple (distance, name).
        """
        if sequence in self._sequences:
            return 0, self._sequences[sequence]
        mindist = len(sequence)
        distances = []
        for seq, name in self._sequences.items():
            ed = edit_distance(seq, sequence, maxdiff=mindist)
            distances.append((ed, name))
            if ed == 1:
                # We know ed does not get smaller because the
                # 'sequence in whitelist' check
                # above covers that
                return ed, name
            mindist = min(mindist, ed)
        distance, name = min(distances)
        return distance, name

    def __len__(self):
        return len(self._sequences)

    def __contains__(self, other):
        return other in self._sequences


def is_chimera(table, whitelist):
    result = pd.Series('', index=table.index, dtype=object)
    if whitelist:
        whitelisted = table[table['whitelist_diff'] == 0]
    else:
        whitelisted = table[table['database_diff'] == 0]
    chimera_finder = ChimeraFinder(list(whitelisted['consensus']))
    for row in table[(table['whitelist_diff'] != 0) & (table['database_diff'] != 0)].itertuples():
        query = row.consensus
        chimera_result = chimera_finder.find_exact(query)
        if chimera_result:
            (prefix_length, prefix_indices, suffix_indices) = chimera_result
            suffix_length = len(query) - prefix_length
            prefix_name = whitelisted.iloc[prefix_indices[0]]['name']
            suffix_row = whitelisted.iloc[suffix_indices[0]]
            suffix_name = suffix_row['name']
            suffix_sequence = suffix_row['consensus']
            # logger.info('Candidate %s (diffs.: %d) appears to be a chimera of '
            #     '%s:1..%d and %s:%d..%d', row.name, row.whitelist_diff, prefix_name, prefix_length,
            #     suffix_name, len(suffix_sequence) - suffix_length + 1, len(suffix_sequence))
            result.loc[row.Index] = '{}:1..{}+{}:{}..{}'.format(prefix_name, prefix_length,
                suffix_name, len(suffix_sequence) - suffix_length + 1, len(suffix_sequence))

    return result


def mark_rows(table, condition, reason):
    table.loc[condition, 'why_filtered'] += reason + ';'
    table.loc[condition, 'is_filtered'] += 1
    logger.info('Marked %s candidates as %r', sum(condition), reason)


def main(args):
    if args.unique_D_threshold <= 1:
        sys.exit('--unique-D-threshold must be at least 1')
    filters = [IdenticalSequenceFilter()]
    if args.cross_mapping_ratio:
        filters.append(CrossMappingFilter(args.cross_mapping_ratio))
    if args.clonotype_ratio:
        filters.append(ClonotypeAlleleRatioFilter(args.clonotype_ratio))
    if args.exact_ratio:
        filters.append(ExactRatioFilter(args.exact_ratio))
    if args.unique_D_ratio or args.unique_D_threshold:
        filters.append(UniqueDRatioFilter(args.unique_D_ratio, args.unique_D_threshold))
    whitelist = Whitelist()
    for path in args.whitelist:
        whitelist.add_fasta(path)
    logger.info('%d unique sequences in whitelist', len(whitelist))

    # Read in tables
    total = 0
    overall_table = None
    for path in args.tables:
        table = pd.read_csv(path, sep='\t')
        i = list(table.columns).index('consensus')
        # whitelist_diff distinguishes between 0 and !=0 only
        # at this point. Accurate edit distances are computed later.
        whitelist_diff = [(0 if s in whitelist else -1) for s in table['consensus']]
        # TODO rename to is_whitelisted
        table.insert(i, 'whitelist_diff', pd.Series(whitelist_diff, index=table.index, dtype=int))
        table.insert(i+1, 'closest_whitelist', pd.Series('', index=table.index))
        table.insert(3, 'why_filtered', pd.Series('', index=table.index))
        table.insert(3, 'is_filtered', pd.Series(0, index=table.index))

        mark_rows(table, table.database_diff < args.minimum_db_diff, 'too_low_dbdiff')

        if 'N_bases' in table.columns:
            mark_rows(table, table.N_bases > args.maximum_N, 'too_many_N_bases')
        mark_rows(table, table.CDR3s_exact < args.unique_cdr3s, 'too_low_CDR3s_exact')
        mark_rows(table, table.CDR3_shared_ratio > args.cdr3_shared_ratio, 'too_high_CDR3_shared_ratio')
        mark_rows(table, table.Js_exact < args.unique_J, 'too_low_Js_exact')
        if not args.allow_stop:
            mark_rows(table, (table.has_stop != 0) & (table.whitelist_diff != 0), 'has_stop')
        mark_rows(table, (table.cluster_size < args.cluster_size) & (table.whitelist_diff != 0),
            'too_low_cluster_size')
        table['database_changes'].fillna('', inplace=True)
        logger.info('Table read from %r contains %s candidate V gene sequences. '
            '%s remain after per-entry filtering', path,
            len(table), sum(table.is_filtered == 0))
        if args.whitelist:
            logger.info('Of those, %d are protected by the whitelist', sum(table.whitelist_diff == 0))
        total += len(table)

        if overall_table is None:
            overall_table = table
        else:
            overall_table.append(table)
    del table
    if len(args.tables) > 1:
        logger.info('Read %s tables with %s entries total. '
            'After per-entry filtering, %s entries remain.', len(args.tables),
            len(overall_table), sum(overall_table.is_filtered == 0))

    filterer = CandidateFilterer(filters)
    filterer.apply(overall_table)

    # Name sequences
    overall_table['name'] = overall_table['name'].apply(UniqueNamer())
    overall_table.sort_values(['name'], inplace=True)

    # Because whitelist_dist() is expensive, this is run when
    # all of the filtering has already been done
    if whitelist:
        for row in overall_table.itertuples():
            # TODO skipping this is just a performance optimization
            if row.is_filtered > 0:
                continue
            distance, name = whitelist.closest(overall_table.loc[row[0], 'consensus'])
            overall_table.loc[row[0], 'closest_whitelist'] = name
            overall_table.loc[row[0], 'whitelist_diff'] = distance
    else:
        overall_table.whitelist_diff.replace(-1, '', inplace=True)

    i = list(overall_table.columns).index('database_diff')
    # TODO
    # chimeras are computed for the full table, not the filtered one. what should be done?
    overall_table.insert(i, 'chimera', is_chimera(overall_table, whitelist))

    # Discard chimeric sequences
    # if not args.allow_chimeras:
    #   overall_table = overall_table[~is_chimera(overall_table, whitelist)].copy()

    filtered_table = overall_table[overall_table.is_filtered == 0]
    del filtered_table['is_filtered']
    del filtered_table['why_filtered']
    print(filtered_table.to_csv(sep='\t', index=False, float_format='%.2f'), end='')

    if args.annotate:
        with open(args.annotate, 'w') as f:
            print(overall_table.to_csv(sep='\t', index=False, float_format='%.2f'), end='', file=f)
    if args.fasta:
        with open(args.fasta, 'w') as f:
            for _, row in filtered_table.iterrows():
                print('>{}\n{}'.format(row['name'], row['consensus']), file=f)

    logger.info('%d sequences in new database', len(filtered_table))



"""
Discover candidate new V genes within a single antibody library

Existing V sequences are grouped by their V gene assignment, and within each
group, consensus sequences are computed.
"""
import csv
import sys
import hashlib
import logging
import random
import itertools
import multiprocessing
from contextlib import ExitStack
from collections import namedtuple, Counter

import dnaio
import numpy as np
import pandas as pd
from tinyalign import edit_distance

from ..table import read_table
from ..cluster import cluster_sequences, single_linkage
from ..utils import (unique_name, downsampled, SerialPool, Merger, has_stop,
    available_cpu_count)
from ..align import iterative_consensus, describe_nt_change

logger = logging.getLogger(__name__)


MINGROUPSIZE = 5
MINEXPRESSED = 10
MAXIMUM_SUBSAMPLE_SIZE = 1600


def add_arguments(parser):
    arg = parser.add_argument
    arg('--threads', '-j', type=int, default=min(4, available_cpu_count()),
        help='Number of threads. Default: no. of available CPUs, but at most 4')
    arg('--seed', type=int, default=None,
        help='Seed value for random numbers for reproducible runs.')
    arg('--consensus-threshold', '-t', metavar='PERCENT', type=float, default=60,
        help='Threshold for consensus computation. Default: %(default)s%%')
    arg('--gene', '-g', action='append', default=[],
        help='Compute consensus for this gene. Can be given multiple times. '
            'Default: Compute for all genes.')
    arg('--limit', type=int, default=None, metavar='N',
        help='Skip remaining genes as soon as at least N candidates were generated. Default: No limit')
    arg('--left', '-l', type=float, metavar='ERROR-RATE',
        help='For consensus, include only sequences that have at least this '
            'error rate (in percent). Default: %(default)s', default=0)
    arg('--right', '-r', type=float, metavar='ERROR-RATE',
        help='For consensus, include only sequences that have at most this '
            'error rate (in percent). Default: %(default)s', default=100)
    arg('--window-width', '-w', type=float, metavar='PERCENT',
        help='Compute consensus for all PERCENT-wide windows. Set to 0 to '
            'disable. Default: %(default)s', default=2)
    arg('--no-cluster', dest='cluster', action='store_false', default=True,
        help='Do not run linkage cluster analysis.')
    arg('--cluster-exact', metavar='N', type=int, default=0,
        help='Treat N exact occurrences of a sequence as a cluster. '
            'Default: Do not cluster exact occurrences')
    arg('--max-n-bases', type=int, default=0, metavar='MAXN',
        help='Remove rows that have more than MAXN "N" nucleotides. If >0, an '
            'N_bases column is added. Default: %(default)s')
    arg('--subsample', metavar='N', type=int, default=1000,
        help='When clustering, use N randomly chosen sequences. Default: %(default)s')
    arg('--ignore-J', action='store_true', default=False,
        help='Include also rows without J assignment or J%%SHM>0.')
    arg('--exact-copies', metavar='N', type=int, default=1,
        help='When subsampling, first pick rows whose V gene sequences'
            'have at least N exact copies in the input. Default: %(default)s')
    arg('--d-evalue', metavar='EVALUE', type=float, default=1E-4,
        help='For Ds_exact, require D matches with an E-value of '
            'at most EVALUE. Default: %(default)s')
    arg('--d-coverage', '--D-coverage', metavar='COVERAGE', type=float, default=70,
        help='For Ds_exact, require D matches with a minimum D '
            'coverage of COVERAGE (in percent). Default: %(default)s)')
    arg('--clonotype-diff', metavar='DIFFERENCES', type=int, default=6,
        help='When clustering CDR3s to computer the no. of clonotypes, allow DIFFERENCES '
            'between (nucleotide-)sequences. Default: %(default)s')

    arg('--table-output', '-o', metavar='DIRECTORY',
        help='Output tables for all analyzed genes to DIRECTORY. '
            'Files will be named <GENE>.tab.')
    arg('--database', metavar='FASTA', default=None,
        help='FASTA file with V genes. If provided, differences between consensus '
            'and database will be computed.')
    arg('--read-names', metavar='FILE',
        help='Write names of reads with exact matches used in discovering each candidate '
            'to FILE')
    arg('table', help='Table with parsed IgBLAST results')


Groupinfo = namedtuple('Groupinfo',
    'count unique_D unique_J unique_CDR3 shared_CDR3_ratio clonotypes read_names unique_barcodes')

SiblingInfo = namedtuple('SiblingInfo', 'sequence requested name group')


def safe_divide(x, y):
    return x / y if y != 0 else 0


class SiblingMerger(Merger):
    """
    Merge very similar consensus sequences into single entries. This could be
    seen as a type of clustering using very specific criteria. Two sequences
    are merged if one is the prefix of the other, allowing differences where
    one of the sequences has an 'N' base.
    """
    def merged(self, s, t):
        chars = []
        for c1, c2 in itertools.zip_longest(s.sequence, t.sequence):
            if c1 is None:
                c = c2
            elif c2 is None:
                c = c1
            elif c1 == 'N':
                c = c2
            elif c2 == 'N':
                c = c1
            elif c1 != c2:
                return None
            else:
                assert c1 == c2
                c = c1
            chars.append(c)
        seq = ''.join(chars)
        requested = s.requested or t.requested
        name = s.name + ';' + t.name
        # take union of groups
        group = pd.concat([s.group, t.group]).groupby(level=0).last()
        return SiblingInfo(seq, requested, name, group)


class Discoverer:
    """
    Discover candidates for novel V genes.
    """
    def __init__(
        self,
        database,
        windows,
        left,
        right,
        cluster,
        cluster_exact,
        table_output,
        consensus_threshold,
        downsample,
        clonotype_differences,
        cluster_subsample_size,
        max_n_bases,
        exact_copies,
        d_coverage,
        d_evalue,
        seed,
        cdr3_counts
    ):
        """
        :param cdr3_counts:
        """
        self.database = database
        self.windows = windows
        self.left = left
        self.right = right
        self.cluster = cluster
        self.cluster_exact = cluster_exact
        self.table_output = table_output
        self.consensus_threshold = consensus_threshold
        self.downsample = downsample
        self.clonotype_differences = clonotype_differences
        self.cluster_subsample_size = cluster_subsample_size
        self.max_n_bases = max_n_bases
        self.exact_copies = exact_copies
        self.d_coverage = d_coverage
        self.d_evalue = d_evalue
        self.seed = seed
        self.cdr3_counts = cdr3_counts

    def _sibling_sequence(self, gene, group):
        """
        For a given group, compute a consensus sequence over the V gene sequences
        in that group.

        If the found sibling is slightly longer or shorter than the version in
        the database, adjust it so it corresponds to the database version exactly.
        """
        sequence = iterative_consensus(list(group.V_nt), program='muscle-medium',
            threshold=self.consensus_threshold/100,
            maximum_subsample_size=self.downsample)
        if gene in self.database:
            database_sequence = self.database[gene]
            if sequence.startswith(database_sequence) or database_sequence.startswith(sequence):
                return database_sequence
        return sequence

    @staticmethod
    def _guess_chain(group):
        """
        Return a guess for the chain type of a given group
        """
        return Counter(group.chain).most_common()[0][0]

    @staticmethod
    def _guess_cdr3_start(group):
        """
        Return a guess for the CDR3 start within sequences in the given group
        """
        return Counter(group.V_CDR3_start).most_common()[0][0]

    def count_unique_d(self, table):
        g = table[(table.D_errors == 0) &
            (table.D_covered >= self.d_coverage) &
            (table.D_evalue <= self.d_evalue)]
        return len(set(s for s in g.D_gene if s))

    @staticmethod
    def count_unique_barcodes(group):
        return len(set(s for s in group.barcode if s))

    def count_clonotypes(self, table):
        """
        Cluster sequences by edit distance and return the number of clusters.

        The sequences are first group by their J assignment and cluster
        numbers are computed within these groups separately, then summed up.
        """
        distance = self.clonotype_differences

        def linked(s, t):
            return edit_distance(s, t, distance) <= distance

        total = 0
        for j_gene, group in table.groupby('J_gene'):
            sequences = list(set(s for s in group.CDR3_nt if s))
            components = single_linkage(sequences, linked)
            total += len(components)
        return total

    def _cluster_siblings(self, gene, table):
        """Find candidates by clustering sequences assigned to one gene"""

        if self.exact_copies > 1:
            # Preferentially pick those sequences (for subsampling) that have
            # multiple exact copies, then fill up with the others
            exact_group = table[table.copies >= self.exact_copies]
            indices = downsampled(list(exact_group.index), self.cluster_subsample_size)
            if len(indices) < self.cluster_subsample_size:
                not_exact_group = table[table.copies < self.exact_copies]
                indices.extend(downsampled(list(not_exact_group.index),
                    self.cluster_subsample_size - len(indices)))
        else:
            indices = downsampled(list(table.index), self.cluster_subsample_size)
        # Ignore CDR3 part of the V sequence for clustering
        sequences_no_cdr3 = list(table.V_no_CDR3.loc[indices])
        df, linkage, clusters = cluster_sequences(sequences_no_cdr3, MINGROUPSIZE)
        n_clusters = len(set(clusters))
        logger.info('%6d %s assignments generated %d cluster%s',
            len(table), gene, n_clusters, 's' if n_clusters != 1 else '')
        cluster_indices = [[] for _ in range(max(clusters) + 1)]
        for i, cluster_id in enumerate(clusters):
            cluster_indices[cluster_id].append(indices[i])

        cl = 0
        for ind in cluster_indices:
            group = table.loc[ind]
            if len(group) < MINGROUPSIZE:
                continue
            sibling = self._sibling_sequence(gene, group)
            cl += 1
            yield SiblingInfo(
                sequence=sibling, requested=False, name='cl{}'.format(cl), group=group)

    def _window_siblings(self, gene, table):
        """
        Find candidates by clustering sequences that have a similar number of differences
        to the reference sequence
        """
        for left, right in self.windows:
            left, right = float(left), float(right)
            group = table[(left <= table.V_SHM) & (table.V_SHM < right)]
            if len(group) < MINGROUPSIZE:
                continue
            sibling = self._sibling_sequence(gene, group)
            if left == int(left):
                left = int(left)
            if right == int(right):
                right = int(right)
            requested = (left, right) == (self.left, self.right)
            if (left, right) == (0, 100):
                name = 'all'
            else:
                name = '{}-{}'.format(left, right)
            yield SiblingInfo(sequence=sibling, requested=requested, name=name, group=group)

    def _cluster_exact_candidates(self, gene, table):
        index = 1
        for sequence, group in table.groupby('V_nt'):
            if len(group) >= self.cluster_exact:
                name = 'ex{}'.format(index)
                index += 1
                yield SiblingInfo(sequence, False, name, group)

    def _collect_siblings(self, gene, group):
        """
        gene -- gene name
        group -- pandas.DataFrame of sequences assigned to that gene

        Yield SiblingInfo objects
        """
        group = group.copy()
        # the original reference sequence for all the IgBLAST assignments in this group
        database_sequence = self.database.get(gene, None)
        database_sequence_found = False

        candidate_iterators = [self._window_siblings(gene, group)]
        if self.cluster:
            candidate_iterators.append(self._cluster_siblings(gene, group))
        if self.cluster_exact:
            candidate_iterators.append(self._cluster_exact_candidates(gene, group))
        for sibling in itertools.chain(*candidate_iterators):
            if sibling.sequence == database_sequence:
                database_sequence_found = True
            yield sibling

        if database_sequence:
            # If this is a database sequence and there are some exact occurrences of it,
            # add it to the list of candidates even if it has not been found as a
            # cluster.
            group_in_window = group[group.V_errors == 0]
            if len(group_in_window) >= MINEXPRESSED:
                if not database_sequence_found:
                    logger.info('Database sequence %r seems to be expressed, but is missing from '
                        'candidates. Re-adding it.', gene)
                yield SiblingInfo(database_sequence, False, 'db', group_in_window)

    def set_random_seed(self, name):
        """Set random seed depending on gene name and seed given to constructor"""
        h = hashlib.md5(name.encode()).digest()[:4]
        n = int.from_bytes(h, byteorder='big')
        random.seed(n + self.seed)

    def __call__(self, args):
        """
        Discover new V genes. args is a tuple (gene, group)
        gene -- name of the gene
        assignments -- a pandas DataFrame with the assignments to the gene
        """
        gene, assignments = args
        self.set_random_seed(gene)
        siblings = SiblingMerger()
        for sibling in self._collect_siblings(gene, assignments):
            siblings.add(sibling)

        candidates = []
        for sibling_info in siblings:
            sibling = sibling_info.sequence
            n_bases = sibling.count('N')
            if n_bases > self.max_n_bases:
                logger.debug('Sibling %s has too many N bases', sibling_info.name)
                continue

            # Sequence without the CDR3-covering part
            sibling_no_cdr3 = sibling[:self._guess_cdr3_start(assignments)]
            group_exact_v = assignments[assignments.V_no_CDR3 == sibling_no_cdr3]
            group_full_exact_v = assignments[assignments['VDJ_nt'].str.startswith(sibling, na=False)]
            groups = (
                ('window', sibling_info.group),
                ('exact', group_exact_v),
                ('full_exact', group_full_exact_v),
            )
            del sibling_no_cdr3

            # self.cdr3_counts are CDR3 counts of all CDR3s in the entire table.
            # We restrict this here to the counts for CDR3s belonging to
            # clusters other than the current one.
            other_cdr3_counts = self.cdr3_counts - Counter(s for s in sibling_info.group.CDR3_nt if s)
            info = dict()
            for key, group in groups:
                cdr3_counts = Counter(s for s in group.CDR3_nt if s)
                unique_cdr3 = len(cdr3_counts)
                shared_cdr3_ratio = safe_divide(len(other_cdr3_counts & cdr3_counts), unique_cdr3)
                unique_j = len(set(s for s in group.J_gene if s))
                clonotypes = self.count_clonotypes(group)
                unique_d = self.count_unique_d(group)
                unique_barcodes = self.count_unique_barcodes(group)
                count = len(group.index)
                read_names = list(group.name)
                info[key] = Groupinfo(count=count, unique_D=unique_d, unique_J=unique_j,
                    unique_CDR3=unique_cdr3, shared_CDR3_ratio=shared_cdr3_ratio,
                    clonotypes=clonotypes, read_names=read_names,
                    unique_barcodes=unique_barcodes)
            if gene in self.database:
                database_diff = edit_distance(sibling, self.database[gene])
                database_changes = describe_nt_change(self.database[gene], sibling)
            else:
                database_diff = None
                database_changes = None

            # Build the Candidate
            sequence_id = gene if database_diff == 0 else unique_name(gene, sibling)
            chain = self._guess_chain(sibling_info.group)
            cdr3_start = self._guess_cdr3_start(sibling_info.group)
            ratio = safe_divide(info['exact'].count, info['exact'].unique_CDR3)

            # Apply some very light filtering on non-database sequences
            if database_diff > 0 and info['exact'].count < 2:
                continue

            candidate = Candidate(
                name=sequence_id,
                source=gene,
                chain=chain,
                cluster=sibling_info.name,
                cluster_size=info['window'].count,
                Js=info['window'].unique_J,
                CDR3s=info['window'].unique_CDR3,
                exact=info['exact'].count,
                full_exact=info['full_exact'].count,
                barcodes_exact=info['exact'].unique_barcodes,
                Ds_exact=info['exact'].unique_D,
                Js_exact=info['exact'].unique_J,
                CDR3s_exact=info['exact'].unique_CDR3,
                clonotypes=info['exact'].clonotypes,
                CDR3_exact_ratio=ratio,
                CDR3_shared_ratio=info['exact'].shared_CDR3_ratio,
                N_bases=n_bases,
                database_diff=database_diff,
                database_changes=database_changes,
                has_stop=has_stop(sibling),
                CDR3_start=cdr3_start,
                consensus=sibling,
                read_names=info['exact'].read_names,
            )
            candidates.append(candidate)
        return candidates


class Candidate(namedtuple('_Candidate', [
    'name',
    'source',
    'chain',
    'cluster',
    'cluster_size',
    'Js',
    'CDR3s',
    'exact',
    'full_exact',
    'barcodes_exact',
    'Ds_exact',
    'Js_exact',
    'CDR3s_exact',
    'clonotypes',
    'CDR3_exact_ratio',
    'CDR3_shared_ratio',
    'N_bases',
    'database_diff',
    'database_changes',
    'has_stop',
    'CDR3_start',
    'consensus',
    'read_names',
])):
    def formatted_dict(self):
        d = self._asdict()
        d['has_stop'] = int(d['has_stop'])
        for name in 'CDR3_exact_ratio', 'CDR3_shared_ratio':
            d[name] = '{:.2f}'.format(d[name])
        del d['read_names']
        return d


def count_prefixes(sequences):
    """
    Count how often each sequence occurs in the given list of
    sequences. If one sequence is the prefix of another one,
    they are considered to be 'identical'.

    Return a dictionary that maps sequence to count.

    >>> r = count_prefixes(['A', 'BB', 'CD', 'CDE', 'CDEF'])
    >>> r == {'A': 1, 'BB': 1, 'CD': 3, 'CDE': 3, 'CDEF': 3}
    True
    """
    sequences = sorted(sequences)
    sequences.append('GUARD')
    prev = 'X'
    start = 0
    count = dict()
    for i, s in enumerate(sequences):
        if not s.startswith(prev):
            # end of a run
            for j in range(start, i):
                count[sequences[j]] = i - start
            start = i
        prev = s
    return count


def main(args):
    if args.database:
        with dnaio.open(args.database) as sr:
            database = {record.name: record.sequence.upper() for record in sr}
    else:
        database = dict()

    if args.seed:
        seed = args.seed
    else:
        seed = random.randrange(10**6)
        logger.info('Use --seed=%d to reproduce this run', seed)

    table = read_table(args.table, usecols=('name', 'chain', 'V_gene', 'D_gene', 'J_gene', 'V_nt',
        'CDR3_nt', 'barcode', 'V_CDR3_start', 'V_SHM', 'J_SHM', 'D_covered', 'D_evalue', 'V_errors',
        'D_errors', 'J_errors', 'VDJ_nt'))
    table['V_no_CDR3'] = [s[:start] if start != 0 else s for s, start in
        zip(table.V_nt, table.V_CDR3_start)]

    logger.info('%s rows read', len(table))
    if not args.ignore_J:
        # Discard rows with any mutation within J at all
        table = table[table.J_SHM == 0][:]
        logger.info('%s rows remain after discarding J%%SHM > 0', len(table))

    if args.exact_copies > 1:
        multiplicities = count_prefixes(table.V_no_CDR3)
        table['copies'] = table.V_no_CDR3.map(multiplicities)
        logger.info('%s rows contain V sequences with at least %s copies',
            sum(table.copies >= args.exact_copies), args.exact_copies)

    columns = list(Candidate._fields)
    if not args.max_n_bases:
        columns.remove('N_bases')
    columns.remove('read_names')
    writer = csv.DictWriter(sys.stdout, fieldnames=columns, delimiter='\t', lineterminator='\n', extrasaction='ignore')
    writer.writeheader()
    genes = set(args.gene)
    if args.window_width:
        windows = [(start, start + args.window_width)
            for start in np.arange(0, 20, args.window_width)]
        logger.info('Using an error rate window of %.1f%% to %.1f%%', args.left, args.right)
        windows.append((args.left, args.right))
    else:
        windows = []

    groups = []
    for gene, group in table.groupby('V_gene'):
        if genes and gene not in genes:
            continue
        if len(group) < MINGROUPSIZE:
            continue
        groups.append((gene, group))

    cdr3_counts = Counter(s for s in table.CDR3_nt if s)
    logger.info('%s unique CDR3s detected overall', len(cdr3_counts))
    discoverer = Discoverer(
        database,
        windows,
        args.left,
        args.right,
        args.cluster,
        args.cluster_exact,
        args.table_output,
        args.consensus_threshold,
        MAXIMUM_SUBSAMPLE_SIZE,
        clonotype_differences=args.clonotype_diff,
        cluster_subsample_size=args.subsample,
        max_n_bases=args.max_n_bases,
        exact_copies=args.exact_copies,
        d_coverage=args.d_coverage,
        d_evalue=args.d_evalue,
        seed=seed,
        cdr3_counts=cdr3_counts
    )

    Pool = SerialPool if args.threads == 1 else multiprocessing.Pool
    n_candidates = 0
    read_names_file = None
    with ExitStack() as stack:
        pool = stack.enter_context(Pool(args.threads))
        if args.read_names:
            read_names_file = stack.enter_context(open(args.read_names, 'w'))
        for candidates in pool.imap(discoverer, groups, chunksize=1):
            for candidate in candidates:
                writer.writerow(candidate.formatted_dict())
                if read_names_file:
                    print(candidate.name, *candidate.read_names, sep='\t', file=read_names_file)
            n_candidates += len(candidates)
            if args.limit is not None and n_candidates >= args.limit:
                break
            sys.stdout.flush()
    logger.info('%s candidate sequences for %s gene(s) generated', n_candidates, len(groups))



import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    sys.stdout.write("At least Python 3.6 is required.\n")
    sys.exit(1)

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="igdiscover",
    setup_requires=["setuptools_scm"],  # Support pip versions that don't know about pyproject.toml
    use_scm_version=True,
    author="Marcel Martin",
    author_email="marcel.martin@scilifelab.se",
    url="https://igdiscover.se/",
    description="Analyze antibody repertoires and discover new V genes",
    long_description=long_description,
    license="MIT",
    entry_points={"console_scripts": ["igdiscover = igdiscover.__main__:main"]},
    package_dir={'': 'src'},
    packages=find_packages('src'),
    package_data={"igdiscover": ["igdiscover.yaml", "Snakefile", "empty.aux"]},
    python_requires=">=3.6",
    install_requires=[
        "pandas>=0.23",
        "numpy",
        "matplotlib>=3.0",
        "seaborn>=0.9",
        "snakemake>=4.5",
        "cutadapt>=2.5",
        "scipy>=1.1",
        "xopen>=0.8.2",
        "ruamel.yaml",
        "dnaio",
        "tinyalign>=0.2",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



"""
Function for reading the table created by the 'parse' subcommand.
"""
import logging
import pandas as pd

logger = logging.getLogger(__name__)


# These columns contain string data
# convert them to str to avoid a PerformanceWarning
# TODO some of these are actually categorical or bool
_STRING_COLUMNS = [
    'V_gene',  # categorical
    'D_gene',  # categorical
    'J_gene',  # categorical
    'chain',  # categorical
    'stop',  # bool
    'productive',  # bool
    'UTR',
    'leader',
    'CDR1_nt',
    'CDR1_aa',
    'CDR2_nt',
    'CDR2_aa',
    'CDR3_nt',
    'CDR3_aa',
    'V_nt',
    'V_aa',
    'V_end',
    'VD_junction',
    'D_region',
    'DJ_junction',
    'J_nt',
    'VDJ_nt',
    'VDJ_aa',
    'name',
    'barcode',
    'race_G',
    'genomic_sequence',
]

_INTEGER_COLUMNS = ('V_errors', 'D_errors', 'J_errors', 'V_CDR3_start')


def fix_columns(df):
    """
    Changes DataFrame in-place
    """
    # Convert all string columns to str to avoid a PerformanceWarning
    for col in _STRING_COLUMNS:
        if col not in df:
            continue
        df[col].fillna('', inplace=True)
        df[col] = df[col].astype('str')
        # Empty strings have been set to NaN by read_csv. Replacing
        # by the empty string avoids problems with groupby, which
        # ignores NaN values.
    # Columns that have any NaN values in them cannot be converted to
    # int due to a numpy limitation.
    for col in _INTEGER_COLUMNS:
        if col not in df.columns:
            continue
        if all(df[col].notnull()):
            df[col] = df[col].astype(int)

    # TODO backwards compatibility
    if 'CDR3_clusters' in df.columns:
        df.rename(columns={'CDR3_clusters': 'clonotypes'}, inplace=True)


def read_table(path, usecols=None, log=False, nrows=None):
    """
    Read in the table created by the parse subcommand (typically named *.tab)
    """
    d = pd.read_csv(path, usecols=usecols, sep='\t', nrows=nrows)
    fix_columns(d)
    if log:
        logger.info('%s rows in input table', len(d))
    return d



"""
Plot a clustermap of all sequences assigned to a gene
"""
import os.path
import logging
from ..table import read_table
from ..utils import downsampled, plural_s
from ..cluster import cluster_sequences

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--minimum-group-size', '-m', metavar='N', type=int, default=200,
        help='Do not plot if there are less than N sequences for a gene. Default: %(default)s')
    arg('--gene', '-g', action='append', default=[],
        help='Plot GENE. Can be given multiple times. Default: Plot all genes.')
    arg('--type', choices=('V', 'D', 'J'), default='V',
        help='Gene type. Default: %(default)s')
    arg('--size', metavar='N', type=int, default=300,
        help='Show at most N sequences (with a matrix of size N x N). Default: %(default)s')
    arg('--ignore-J', action='store_true', default=False,
        help='Include also rows without J assignment or J%%SHM>0.')
    arg('--dpi', type=int, default=200, help='Resolution of output file. Default: %(default)s')
    arg('--no-title', dest='title', action='store_false', default=True,
        help='Do not add a title to the plot')
    arg('table', help='Table with parsed and filtered IgBLAST results')
    arg('directory', help='Save clustermaps as PNG into this directory', default=None)


def plot_clustermap(sequences, title, plotpath, size=300, dpi=200):
    """
    Plot a clustermap of the given sequences

    size -- Downsample to this many sequences
    title -- plot title

    Return the number of clusters.
    """
    import seaborn as sns

    logger.info('Clustering %d sequences (downsampled to at most %d)', len(sequences), size)
    sequences = downsampled(sequences, size)
    df, linkage, clusters = cluster_sequences(sequences)

    palette = sns.color_palette([(0.15, 0.15, 0.15)])
    palette += sns.color_palette('Spectral', n_colors=max(clusters), desat=0.9)
    row_colors = [palette[cluster_id] for cluster_id in clusters]
    cm = sns.clustermap(df,
            row_linkage=linkage,
            col_linkage=linkage,
            row_colors=row_colors,
            linewidths=None,
            linecolor='none',
            figsize=(210/25.4, 210/25.4),
            cmap='Blues',
            xticklabels=False,
            yticklabels=False
    )
    if title is not None:
        cm.fig.suptitle(title)
    cm.savefig(plotpath, dpi=dpi)

    # free the memory used by the plot
    import matplotlib.pyplot as plt
    plt.close('all')

    return len(set(clusters))


def main(args):
    import matplotlib
    matplotlib.use('agg')
    import seaborn as sns
    sns.set()

    if not os.path.exists(args.directory):
        os.mkdir(args.directory)
    gene_col = args.type + '_gene'
    seq_col = args.type + '_nt'
    usecols = ['J_SHM', 'V_gene', gene_col, seq_col]
    table = read_table(args.table, usecols=usecols)

    # Discard rows with any mutation within J at all
    logger.info('%s rows read', len(table))
    if not args.ignore_J:
        # Discard rows with any mutation within J at all
        table = table[table.J_SHM == 0][:]
        logger.info('%s rows remain after discarding J%%SHM > 0', len(table))

    path_sanitizer = {ord(c): None for c in "\\/*?[]=\"'"}
    genes = frozenset(args.gene)
    n = 0
    too_few = 0
    for gene, group in table.groupby(gene_col):
        if genes and gene not in genes:
            continue
        if len(group) < args.minimum_group_size:
            too_few += 1
            continue
        title = gene if args.title else None
        path = os.path.join(args.directory, gene.translate(path_sanitizer) + '.png')
        sequences = list(group[seq_col])
        n_clusters = plot_clustermap(sequences, title, path, size=args.size, dpi=args.dpi)
        n += 1
        logger.info('Plotted %r with %d cluster%s', gene, n_clusters, plural_s(n_clusters))
        #for i, cons in enumerate(consensus_sequences):
            #print('>{}_cluster{}\n{}'.format(gene, ('red', 'blue')[i], cons))
            #print('number of Ns:', cons.count('N'))
        #if len(consensus_sequences) >= 2:
            #print('difference between consensuses:', edit_distance(*consensus_sequences[:2]))

    logger.info('%s plots created (%s skipped because of too few sequences)', n, too_few)



"""
Plot allele usage
"""
import sys
import logging
import pandas as pd
import dnaio

from ..table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--d-evalue', type=float, default=1E-4,
        help='Maximal allowed E-value for D gene match. Default: %(default)s')
    arg('--d-coverage', '--D-coverage', type=float, default=65,
        help='Minimum D coverage (in percent). Default: %(default)s%%)')
    arg('--database', metavar='FASTA',
        help='Restrict plotting to the sequences named in the FASTA file. '
             'Only the sequence names are used!')
    arg('--order', metavar='FASTA',
        help='Sort genes according to the order of the records in the FASTA file.')
    arg('--x', choices=('V', 'D', 'J'), default='V',
        help='Type of gene on x axis. Default: %(default)s')
    arg('--gene', choices=('V', 'D', 'J'), default='J',
        help='Type of gene on y axis. Default: %(default)s')
    arg('alleles', help='List of alleles to plot on y axis, separated by comma')
    arg('table', help='Table with parsed and filtered IgBLAST results')
    arg('plot', help='Path to output PDF or PNG')


def main(args):
    usecols = ['V_gene', 'D_gene', 'J_gene', 'V_errors', 'D_errors', 'J_errors', 'D_covered',
        'D_evalue']

    # Support reading a table without D_errors
    try:
        table = read_table(args.table, usecols=usecols)
    except ValueError:
        usecols.remove('D_errors')
        table = read_table(args.table, usecols=usecols)
    logger.info('Table with %s rows read', len(table))
    if args.x == 'V' or args.gene == 'V':
        table = table[table.V_errors == 0]
        logger.info('%s rows remain after requiring V errors = 0', len(table))
    if args.gene == 'J' or args.x == 'J':
        table = table[table.J_errors == 0]
        logger.info('%s rows remain after requiring J errors = 0', len(table))
    if args.gene == 'D' or args.x == 'D':
        table = table[table.D_evalue <= args.d_evalue]
        logger.info('%s rows remain after requiring D E-value <= %s', len(table), args.d_evalue)
        table = table[table.D_covered >= args.d_coverage]
        logger.info('%s rows remain after requiring D coverage >= %s', len(table), args.d_coverage)
        if 'D_errors' in table.columns:
            table = table[table.D_errors == 0]
            logger.info('%s rows remain after requiring D errors = 0', len(table))

    gene1 = args.x + '_gene'
    gene2 = args.gene + '_gene'
    expression_counts = table.groupby((gene1, gene2)).size().to_frame().reset_index()
    matrix = pd.DataFrame(
        expression_counts.pivot(index=gene1, columns=gene2, values=0).fillna(0), dtype=int)
    # matrix[v_gene,d_gene] gives co-occurrences of v_gene and d_gene
    print('#\n# Expressed genes with counts\n#')
    # The .sum() is along axis=0, that is, the V gene counts are summed up,
    # resulting in counts for each D/J gene
    for g, count in matrix.sum().iteritems():
        print(g, '{:8}'.format(count))

    alleles = args.alleles.split(',')
    for allele in alleles:
        if allele not in matrix.columns:
            logger.error('Allele %s not expressed in this dataset', allele)
            sys.exit(1)

    matrix = matrix.loc[:, alleles]

    if args.database:
        with dnaio.open(args.database) as f:
            x_names = [record.name for record in f if record.name in matrix.index]
        if not x_names:
            logger.error('None of the sequence names in %r were found in the input table',
                args.database)
            sys.exit(1)
        matrix = matrix.loc[x_names, :]

    if args.order:
        with dnaio.open(args.order) as f:
            ordered_names = [r.name.partition('*')[0] for r in f]
        gene_order = {name: index for index, name in enumerate(ordered_names)}

        def orderfunc(full_name):
            name, _, allele = full_name.partition('*')
            allele = int(allele)
            try:
                index = gene_order[name]
            except KeyError:
                logger.warning('Gene name %s not found in %r, placing it at the end',
                    name, args.order)
                index = 1000000
            return index * 1000 + allele
        matrix['V_gene_tmp'] = pd.Series(matrix.index, index=matrix.index).apply(orderfunc)
        matrix.sort_values('V_gene_tmp', inplace=True)
        del matrix['V_gene_tmp']

    print('#\n# Allele-specific expression\n#')
    print(matrix)

    if len(alleles) == 2:
        matrix.loc[:, alleles[1]] *= -1

    # remove all-zero rows
    # allele_expressions = allele_expressions[(allele_expressions > 0.001).any(axis=1)]
    ax = matrix.plot(kind='bar', stacked=True, figsize=(12, 6))
    ax.legend(title=None)
    ax.set_title('Allele-specific expression counts')
    ax.set_xlabel(args.x + ' gene')
    ax.set_ylabel('Count')
    ax.figure.set_tight_layout(True)
    # ax.legend(bbox_to_anchor=(1.15, 0.5))
    ax.figure.savefig(args.plot)
    logger.info('Plotted %r', args.plot)



"""
Merge paired-end reads (a wrapper around PEAR)

This script can also manage a cache of already-merged files.
"""
import os
import sys
from pathlib import Path
import logging
import tempfile
import subprocess
import hashlib
import shutil

from ..config import GlobalConfig

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--threads', '-j', default=None, type=int, help='Number of threads')
    arg('--no-cache', default=None, action='store_true',
        help='Disable cache. Default: Determined by configuration')
    arg('reads1', help='Forward reads FASTQ file')
    arg('reads2', help='Reverse reads FASTQ file')
    arg('output',  help='Output file (compressed FASTQ)')


def compute_hash(path1: str, path2: str):
    result = subprocess.run(['pear', '-h'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    pear_version = None
    for line in result.stdout.split(b'\n'):
        if line.startswith(b'PEAR'):
            pear_version = line
            break
    assert pear_version is not None

    hasher = hashlib.md5(pear_version)
    for path in path1, path2:
        with open(path, 'rb') as f:
            while True:
                chunk = f.read(1048576)
                if not chunk:
                    break
                hasher.update(chunk)
    return hasher.hexdigest()


def run_pear(path1: str, path2: str, output: str, log_output=None, threads: int=None):
    gzip = 'pigz' if shutil.which('pigz') else 'gzip'
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)
        args = ['pear']
        if threads is not None:
            args += ['-j', str(threads)]
        args += [
            '-f', path1,
            '-r', path2,
            '-o', tmpdir / 'merged'
        ]
        if log_output is None:
            subprocess.run(args, check=True)
        else:
            result = subprocess.run(args, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            with open(log_output, 'wb') as f:
                f.write(result.stdout)

        subprocess.run([gzip, tmpdir / 'merged.assembled.fastq'], check=True)
        shutil.move(tmpdir / 'merged.assembled.fastq.gz', output)


def run_pear_cached(path1: str, path2: str, output: str, threads: int = None):
    cache_home = Path(os.environ.get('XDG_CACHE_HOME', os.path.expanduser('~/.cache')))
    cachedir = cache_home / 'igdiscover' / 'mergedreads'
    cachedir.mkdir(parents=True, exist_ok=True)

    md5sum = compute_hash(path1, path2)
    cached_merged = cachedir / (md5sum + '.fastq.gz')
    cached_log = cachedir / (md5sum + '.log')
    if not cached_merged.exists() or not cached_log.exists():
        run_pear(path1, path2, output=cached_merged, log_output=cached_log, threads=threads)
        logger.info('PEAR result copied to cache')
    else:
        logger.info('PEAR result found in cache:')
        logger.info(cached_merged)
        logger.info(cached_log)
    shutil.copy(cached_merged, output)
    with open(cached_log) as f:
        sys.stdout.write(f.read())

    # Update mtime of the cache file. Let’s us find unused cache entries.
    for path in cached_merged, cached_log:
        try:
            path.touch()
        except PermissionError:
            pass


def main(args):
    use_cache = GlobalConfig().use_cache
    if args.no_cache:
        use_cache = False
    if use_cache:
        logger.info('Cache enabled')
        func = run_pear_cached
    else:
        func = run_pear
    func(path1=args.reads1, path2=args.reads2, output=args.output, threads=args.threads)



#!/usr/bin/env python3
"""
Split sequences in a FASTA file by barcode.
"""
__author__ = "Marcel Martin"

import sys
import random
from sqt import HelpfulArgumentParser, FastaReader


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument('--barcode-length', '-l', metavar='N', type=int, default=12,
		help="Barcode length (default: %(default)s)")
	parser.add_argument('--sequential', '-s', action='store_true', default=False,
		help="Number output files sequentially, do not use BARCODE in the file name.")
	parser.add_argument("fasta", metavar='FASTA', help="Input FASTA file")
	parser.add_argument("prefix", metavar='PREFIX', nargs='?', default='split',
		help='Prefix for output file names. The output files will be named '
		'<PREFIX><BARCODE>.fasta or <PREFIX><NUMBER>.fasta.')
	args = parser.parse_args()
	n = 0
	barcodes = dict()
	n_barcodes = 1
	for record in FastaReader(args.fasta):
		barcode = record.sequence[:args.barcode_length]
		if barcode in barcodes:
			index = barcodes[barcode]
		else:
			index = n_barcodes
			barcodes[barcode] = index
			n_barcodes += 1
		if args.sequential:
			path = args.prefix + str(index) + '.fasta'
		else:
			path = args.prefix + barcode + '.fasta'
		# Open/close files every time to avoid too many open files
		with open(path, mode='a') as f:
			f.write('>{}\n{}\n'.format(record.name, record.sequence))
		n += 1
	print('Wrote {} sequences to {} files.'.format(n, len(barcodes)))


if __name__ == '__main__':
        main()



"""
Discover D and J genes

The most frequent D/J sequences are considered candidates.

For J genes, candidate sequences are merged if they overlap each other.

The result table is written to standard output. Use --fasta to also
generate FASTA output.
"""
import logging
import pandas as pd
from collections import defaultdict
from typing import List
import dnaio

from tinyalign import edit_distance
from ..utils import Merger, merge_overlapping, unique_name, is_same_gene, slice_arg
from ..table import read_table, fix_columns


logger = logging.getLogger(__name__)

MINIMUM_CANDIDATE_LENGTH = 5


def add_arguments(parser):
    arg = parser.add_argument
    arg('--database', metavar='FASTA',
        help='FASTA file with reference gene sequences')
    arg('--merge', default=None, action='store_true', help='Merge overlapping genes. '
        'Default: Enabled for D, disabled for J and V.')
    arg('--no-merge', dest='merge', action='store_false', help='Do not merge overlapping genes')
    arg('--gene', default='J', choices=('V', 'D', 'J'), type=str.upper,
        help='Which gene category to discover. Default: %(default)s')
    arg('--j-coverage', type=float, default=None, metavar='PERCENT',
        help='Require that the sequence covers at least PERCENT of the J gene. '
        'Default: 90 when --gene=J; 0 otherwise')
    arg('--allele-ratio', type=float, metavar='RATIO', default=0.2,
        help='Required allele ratio. Works only for genes named "NAME*ALLELE". Default: %(default)s')
    arg('--cross-mapping-ratio', type=float, metavar='RATIO', default=0.1,
        help='Ratio for detection of cross-mapping artifacts. Default: %(default)s')
    arg('--min-count', metavar='N', type=int, default=None,
        help='Omit candidates with fewer than N exact occurrences in the input table. '
            'Default: 1 for J; 10 for D; 100 for V')
    arg('--no-perfect-matches', dest='perfect_matches', default=True, action='store_false',
        help='Do not filter out sequences for which the V assignment (or J for --gene=V) '
            'has at least one error')

    # --gene=D options
    arg('--d-core-length', metavar='L', type=int, default=6,
        help='Use only D core regions that have at least length L (only '
            'applies when --gene=D). Default: %(default)s')
    arg('--d-core', type=slice_arg, default=slice(2, -2),
        help='D core region location (only applies when --gene=D). '
            'Default: %(default)s')

    arg('--fasta', help='Write discovered sequences to FASTA file')
    arg('table', help='Table with parsed and filtered IgBLAST results')


class Candidate:
    __slots__ = ('name', 'sequence', 'exact_occ', 'max_count', 'other_genes', 'db_name',
        'db_distance', 'cdr3s', 'missing')

    def __init__(self, name, sequence, exact_occ=0, max_count=0, cdr3s=None, other_genes=None,
            db_name=None, db_distance=None):
        self.name = name
        self.sequence = sequence
        self.exact_occ = exact_occ
        self.max_count = max_count  # an upper bound for the exact_occ
        self.cdr3s = cdr3s if cdr3s is not None else set()
        self.other_genes = other_genes if other_genes is not None else set()
        self.db_name = db_name
        self.db_distance = db_distance
        self.missing = ''

    @property
    def unique_CDR3(self):
        return len(self.cdr3s)

    def __repr__(self):
        return 'Candidate({sequence!r}, exact_occ={exact_occ}, max_count={max_count}, ...)'.format(
            sequence=self.sequence,
            exact_occ=self.exact_occ,
            max_count=self.max_count,
        )


class OverlappingSequenceMerger(Merger):
    """
    Merge sequences that overlap
    """
    def merged(self, s, t):
        """
        Merge two sequences if they overlap. If they should not be merged,
        None is returned.
        """
        m = merge_overlapping(s.sequence, t.sequence)
        if m is not None:
            return Candidate(s.name, m, max_count=t.max_count + s.max_count)

        return None


class AlleleRatioMerger(Merger):
    """
    Discard sequences with too low allele ratio
    """
    def __init__(self, allele_ratio, cross_mapping_ratio):
        super().__init__()
        self._allele_ratio = allele_ratio
        self._cross_mapping_ratio = cross_mapping_ratio

    def merged(self, s, t):
        """
        Merge two sequences if they overlap. If they should not be merged,
        None is returned.
        """
        # TODO copy-and-pasted from germlinefilter
        #
        # Check allele ratio. Somewhat similar to cross-mapping, but
        # this uses sequence names to decide whether two genes can be
        # alleles of each other and the ratio is between the CDR3s_exact
        # values
        if self._allele_ratio and is_same_gene(s.name, t.name):
            for u, v in [(s, t), (t, s)]:
                if v.unique_CDR3 == 0:
                    continue
                ratio = u.unique_CDR3 / v.unique_CDR3
                if ratio < self._allele_ratio:
                    # logger.info('Allele ratio %.4f too low for %r compared to %r',
                    #     ratio, u.name, v.name)
                    return v

        if self._cross_mapping_ratio:
            # When checking for cross mapping, ignore overhanging bases in the 5' end.
            # Example:
            # ---ACTACGACTA...
            # XXX|||||X||||
            # ATTACTACTACTA...
            if len(t.sequence) < len(s.sequence):
                t, s = s, t  # s is now the shorter sequence
            t_seq = t.sequence[len(t.sequence) - len(s.sequence):]
            s_seq = s.sequence
            dist = edit_distance(s_seq, t_seq, 1)
            if dist > 1:
                return None
            total_occ = (s.exact_occ + t.exact_occ)
            if total_occ == 0:
                return None
            for u, v in [(s, t), (t, s)]:
                ratio = u.exact_occ / total_occ
                if ratio < self._cross_mapping_ratio:
                    # u is probably a cross-mapping artifact of the higher-expressed v
                    logger.info('%r is a cross-mapping artifact of %r (ratio %.4f)',
                        u.name, v.name, ratio)
                    return v

        return None


def filter_by_allele_ratio(table, allele_ratio):
    arm = AlleleRatioMerger(allele_ratio, cross_mapping_ratio=None)
    renamed_counts = table.reset_index().rename(columns={'gene': 'name'})
    arm.extend(renamed_counts.itertuples(index=False))
    counts = pd.DataFrame(list(arm), columns=renamed_counts.columns) \
        .rename(columns={'name': 'gene'}) \
        .set_index('gene')
    return counts


def count_occurrences(candidates, table_path, search_columns, other_gene, other_errors, merge,
        perfect_matches):
    """
    Count how often each candidate sequence occurs in the input table.
    The columns named in search_columns are concatenated and searched.
    This circumvents inaccurate IgBLAST alignment boundaries.
    Rows where the column named by other_errors is not zero are ignored.

    The input table is read in chunks to reduce memory usage.

    candidates -- list of candidates
    table_path -- path to input table
    search_columns -- which columns to search. The contained strings are
        concatenated and then searched.
    merge -- If True, stop searching for other candidates in a single row
        after one candidate has been found.

    The following attributes of the candidates are updated:

    - exact_occ
    - other_genes
    - cdr3s

    Return the updated list of candidates.
    """
    candidates_map = {c.sequence: c for c in candidates}

    search_order = [c.sequence for c in candidates]
    cols = [other_gene, 'V_errors', 'J_errors', 'CDR3_nt'] + search_columns

    search_cache = defaultdict(list)  # map haystack sequence to list of candidates that occur in it
    for chunk in pd.read_csv(table_path, usecols=cols, chunksize=10000, sep='\t'):
        fix_columns(chunk)
        if perfect_matches:
            chunk = chunk[chunk[other_errors] == 0]
        # concatenate search columns
        if len(chunk) == 0:  # TODO that this is needed is possibly a pandas bug
            continue
        chunk['haystack'] = chunk.loc[:, search_columns].astype(str).sum(axis=1)
        chunk['haystack'] = chunk['haystack'].str.replace('(', '').replace(')', '')

        for row in chunk.itertuples():
            if row.haystack not in search_cache:
                for needle in search_order:
                    if needle in row.haystack:
                        search_cache[row.haystack].append(candidates_map[needle])
            for candidate in search_cache[row.haystack]:
                candidate.exact_occ += 1  # TODO += row.count?
                candidate.other_genes.add(getattr(row, other_gene))
                candidate.cdr3s.add(row.CDR3_nt)
                if merge:
                    # When overlapping candidates have been merged,
                    # there will be no other pattern that is a
                    # substring of the current search pattern.
                    break
    return candidates_map.values()


def discard_substring_occurrences(candidates):
    """
    Filter a candidate list by discarding candidates whose sequences are
    substrings of another candidate’s sequence
    """
    # Shorter sequences first
    candidates = sorted(candidates, key=lambda c: len(c.sequence))
    for i, short in enumerate(candidates):
        for long in candidates[i+1:]:
            if short.sequence in long.sequence:
                break
        else:
            # no substring occurrence - keep this candidate
            yield short


def sequence_candidates(table, column, minimum_length, core=slice(None, None), min_occ=3):
    """
    Generate candidates by clustering all sequences in a column
    (V_nt, D_nt or J_nt). At least min_occ occurrences are required.

    core -- a slice object. If given, the strings in the column are
           sliced before being clustered.
    """
    for sequence, occ in table[column].str[core].value_counts().items():
        if len(sequence) >= minimum_length and occ >= min_occ:
            yield Candidate(None, sequence, max_count=occ)


def count_unique_cdr3(table):
    return len(set(s for s in table.CDR3_nt if s))


def count_unique_gene(table, gene_type):
    return len(set(s for s in table[gene_type + '_gene'] if s))


def compute_expressions(table, gene_type):
    """Compute expression counts of known genes"""
    assert gene_type in {'V', 'D', 'J'}
    columns = ('gene', 'count', 'unique_CDR3')
    for gt in 'V', 'D', 'J':
        if gene_type != gt:
            columns += ('unique_' + gt,)
    gene_column = gene_type + '_gene'
    rows = []
    for gene, group in table.groupby(gene_column):
        if gene == '':
            continue
        unique_cdr3 = count_unique_cdr3(group)
        row = dict(gene=gene, count=len(group), unique_CDR3=unique_cdr3)
        for gt in 'V', 'D', 'J':
            if gene_type != gt:
                row['unique_' + gt] = count_unique_gene(group, gene_type=gt)
        rows.append(row)
    counts = pd.DataFrame(rows, columns=columns).set_index('gene')
    return counts


def make_whitelist(table, database, gene_type: str, allele_ratio=None) -> List[str]:
    """
    Return a list of sequences that represent expressed alleles
    """
    assert gene_type in {'V', 'D', 'J'}
    # Compute expression counts in the same way the 'igdiscover count' command would do it
    counts = compute_expressions(table, gene_type)
    if allele_ratio:
        counts = filter_by_allele_ratio(counts, allele_ratio)
    names = list(counts.index)

    # Construct whitelist from all expressed alleles
    database = {r.name: r.sequence for r in database}
    sequences = [database[name] for name in names]

    return sequences


def print_table(candidates, other_gene, missing):
    columns = ['name', 'exact_occ', other_gene + 's', 'CDR3s', 'database', 'database_diff', 'sequence']
    if missing:
        columns.append('missing')
    print(*columns, sep='\t')
    for candidate in candidates:
        columns = [
            candidate.name,
            candidate.exact_occ,
            len(set(candidate.other_genes)),
            candidate.unique_CDR3,
            candidate.db_name if candidate.db_name is not None else '',
            candidate.db_distance if candidate.db_distance is not None else -1,
            candidate.sequence
        ]
        if missing:
            columns.append(candidate.missing)
        print(*columns, sep='\t')


def main(args):
    if args.database:
        with dnaio.open(args.database) as fr:
            database = list(fr)
        logger.info('Read %d sequences from %r', len(database), args.database)
    else:
        database = None
    column = {'V': 'V_nt', 'J': 'J_nt', 'D': 'D_region'}[args.gene]
    other = 'V' if args.gene in ('D', 'J') else 'J'
    other_gene = other + '_gene'
    other_errors = other + '_errors'
    table = read_table(args.table,
        usecols=['count', 'V_gene', 'D_gene', 'J_gene', 'V_errors', 'J_errors', 'J_covered', column, 'CDR3_nt'])
    logger.info('Table with %s rows read', len(table))

    if args.j_coverage is None and args.gene == 'J':
        args.j_coverage = 90
    if args.j_coverage:
        table = table[table['J_covered'] >= args.j_coverage]
        logger.info('Keeping %s rows that have J_covered >= %s', len(table), args.j_coverage)
    if args.perfect_matches:
        table = table[table[other_errors] == 0]
        logger.info('Keeping %s rows that have no %s mismatches', len(table), other)

    if args.merge is None:
        args.merge = args.gene == 'D'
    if args.min_count is None:
        args.min_count = {'J': 1, 'D': 10, 'V': 100}[args.gene]  # TODO J is fine, but are D and V?

    if args.gene == 'D':
        candidates = sequence_candidates(
            table, column, minimum_length=args.d_core_length, core=args.d_core)
    elif args.gene == 'J':
        candidates = sequence_candidates(
            table, column, minimum_length=MINIMUM_CANDIDATE_LENGTH)
    else:
        candidates = sequence_candidates(
            table, column, minimum_length=MINIMUM_CANDIDATE_LENGTH)

    candidates = list(candidates)
    logger.info('Collected %s unique %s sequences', len(candidates), args.gene)

    # Add whitelisted sequences
    if database:
        whitelist = make_whitelist(table, database, args.gene, args.allele_ratio)
        missing_whitelisted = set(whitelist) - set(c.sequence for c in candidates)
        for sequence in missing_whitelisted:
            candidates.append(Candidate(None, sequence))
        logger.info('Added %d whitelisted sequence%s',
            len(missing_whitelisted), 's' if len(missing_whitelisted) != 1 else '')

    candidates = list(discard_substring_occurrences(candidates))
    logger.info('Removing candidate sequences that occur within others results in %s candidates',
        len(candidates))
    candidates = [candidate for candidate in candidates if 'N' not in candidate.sequence]
    logger.info('Removing candidates containing "N" results in %s candidates',
        len(candidates))

    if args.merge:
        logger.info('Merging overlapping sequences ...')
        # Merge candidate sequences that overlap. If one candidate is longer than
        # another, this is typically a sign that IgBLAST has not extended the
        # alignment long enough.
        merger = OverlappingSequenceMerger()
        for candidate in candidates:
            merger.add(candidate)
        logger.info('After merging overlapping %s sequences, %s remain', args.gene, len(merger))
        candidates = list(merger)
    del table

    logger.info('%d candidates', len(candidates))
    # Assign names etc.
    if database:
        for candidate in candidates:
            distances = [(edit_distance(db.sequence, candidate.sequence), db) for db in database]
            candidate.db_distance, closest = min(distances, key=lambda x: x[0])
            candidate.db_name = closest.name

            if candidate.db_distance == 0:
                candidate.name = closest.name
            else:
                # Exact db sequence not found, is there one that contains
                # this candidate as a substring?
                for db_record in database:
                    index = db_record.sequence.find(candidate.sequence)
                    if index == -1:
                        continue
                    if args.gene == 'D':
                        start = db_record.sequence.find(candidate.sequence)
                        prefix = db_record.sequence[:start]
                        suffix = db_record.sequence[start + len(candidate.sequence):]
                        candidate.missing = '{}...{}'.format(prefix, suffix)
                    else:
                        # Replace this record with the full-length version
                        candidate.sequence = db_record.sequence
                        candidate.db_distance = 0
                    candidate.name = db_record.name
                    break
                else:
                    candidate.name = unique_name(closest.name, candidate.sequence)
    else:
        for candidate in candidates:
            candidate.name = unique_name(args.gene, candidate.sequence)

    logger.info('Counting occurrences ...')
    if args.gene == 'D':
        search_columns = ['VD_junction', 'D_region', 'DJ_junction']
    elif args.gene == 'J':
        search_columns = ['DJ_junction', 'J_nt']
    else:
        search_columns = ['genomic_sequence']
    candidates = count_occurrences(candidates, args.table, search_columns, other_gene, other_errors,
        args.merge, args.perfect_matches)

    # Filter by allele ratio
    if args.allele_ratio or args.cross_mapping_ratio:
        arm = AlleleRatioMerger(args.allele_ratio, args.cross_mapping_ratio)
        arm.extend(candidates)
        candidates = list(arm)
        logger.info('After filtering by allele ratio and/or cross-mapping ratio, %d candidates remain',
            len(candidates))

    candidates = sorted(candidates, key=lambda c: c.name)
    candidates = [c for c in candidates if c.exact_occ >= args.min_count or c.db_distance == 0]
    print_table(candidates, other_gene, missing=args.gene == 'D')

    if args.fasta:
        with open(args.fasta, 'w') as f:
            for candidate in sorted(candidates, key=lambda r: r.name):
                print('>{}\n{}'.format(candidate.name, candidate.sequence), file=f)

    logger.info('Wrote %d genes', len(candidates))



"""
Create and initialize a new analysis directory.
"""
import glob
import logging
import re
import os
import os.path
import sys
import subprocess
import pkg_resources
import dnaio
from ..config import Config

try:
    import tkinter as tk
    from tkinter import messagebox
    from tkinter import filedialog
except ImportError:
    tk = None

from xopen import xopen

logger = logging.getLogger(__name__)

do_not_show_cpustats = 1


def add_arguments(parser):
    parser.add_argument('--database', '--db', metavar='PATH', default=None,
        help='Directory with V.fasta, D.fasta and J.fasta files. If not given, a dialog is shown.')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--single-reads', default=None, metavar='READS',
        help='File with single-end reads (.fasta.gz or .fastq.gz)')
    group.add_argument('--reads1', default=None,
        help='First paired-end read file. The second is found automatically. '
            'Must be a .fastq.gz file. If not given, a dialog is shown.')
    parser.add_argument('directory', help='New analysis directory to create')


def launch(path):
    if hasattr(os, 'startfile'):
        os.startfile(path)
    elif sys.platform == 'linux':
        subprocess.call(['xdg-open', path])
    elif sys.platform == 'darwin':
        subprocess.call(['open', path])


class TkinterGui:
    """Show a GUI for selecting reads and the database directory"""
    def __init__(self):
        if not tk:  # import failed
            raise ImportError()
        root = tk.Tk()
        root.withdraw()

    def yesno(self, title, question):
        return tk.messagebox.askyesno(title, question)

    def database_path(self, initialdir):
        path = tk.filedialog.askdirectory(
            title="Choose V/D/J database directory", mustexist=True,
            initialdir=initialdir)
        return path

    def reads1_path(self):
        path = tk.filedialog.askopenfilename(
            title="Choose first reads file",
            filetypes=[
                ("Reads", "*.fastq.gz"),
                ("Any file", "*")])
        return path

    def single_reads_path(self):
        path = tk.filedialog.askopenfilename(
            title="Choose single-end reads file",
            filetypes=[
                ("Reads", "*.fasta *.fastq *.fastq.gz *.fasta.gz"),
                ("Any file", "*")])
        return path

    def error(self, title, message):
        tk.messagebox.showerror(title, message)


"""
# Works, but let’s not introduce the PySide dependency for now.

def qt_path():
    import PySide
    from PySide.QtGui import QApplication
    from PySide.QtGui import QMessageBox, QFileDialog

    # Create the application object
    app = QApplication([])

    path = QFileDialog.getOpenFileName(None,
        "Open first reads file", '.', "FASTA/FASTQ reads (*.fastq *.fasta *.fastq.gz *.fasta.gz);; Any file (*)")
    # QMessageBox.information(None, 'Chosen file', path[0])
    return path[0]
"""


def is_1_2(s, t):
    """
    Determine whether s and t are identical except for a single character of
    which one of them is '1' and the other is '2'.
    """
    differences = 0
    one_two = {'1', '2'}
    for c1, c2 in zip(s, t):
        if c1 != c2:
            differences += 1
            if differences == 2:
                return False
            if {c1, c2} != one_two:
                return False
    return differences == 1


def guess_paired_path(path):
    """
    Given the path to a file that contains the sequences for the first read in a
    pair, return the file that contains the sequences for the second read in a
    pair. Both files must have identical names, except that the first must have
    a '1' in its name, and the second must have a '2' at the same position.

    Return None if no second file was found or if there are too many candidates.

    >>> guess_paired_path('file.1.fastq.gz')  # doctest: +SKIP
    'file.2.fastq.gz'  # if that file exists
    """
    base, name = os.path.split(path)
    # All lone 1 digits replaced with '?'
    name_with_globs = re.sub(r'(?<![0-9])1(?![0-9])', '?', name)
    glob_pattern = os.path.join(base, name_with_globs)
    paths = [p for p in glob.glob(glob_pattern) if is_1_2(p, path)]
    if len(paths) == 1:
        return paths[0]
    return None


class UnknownFileFormatError(Exception):
    pass


def file_type(path):
    """
    Return 'fasta' or 'fastq' depending on file format. The file may optionally
    be gzip-compressed.
    """
    if path.endswith('.gz'):
        file = xopen(path)
    else:
        file = open(path)
    with file as f:
        first_char = f.read(1)
        if first_char == '@':
            return 'fastq'
        elif first_char == '>':
            return 'fasta'
        else:
            raise UnknownFileFormatError('Cannot recognize format. File starts with neither ">" nor "@"')


def try_open(path):
    try:
        with open(path) as f:
            pass
    except OSError as e:
        logger.error('Could not open %r: %s', path, e)
        sys.exit(1)


def read_and_repair_fasta(path):
    """
    Read a FASTA file and make sure it is suitable for use with makeblastdb.
    It repairs the following issues:

    - If a record is empty, it is skipped
    - If a record name occurs more than once, the second record name gets a suffix
    - If a sequence occurs more than once, occurrences after the first are skipped
    """
    with dnaio.open(path) as sr:
        records = list(sr)

    names = set()
    sequences = dict()
    for r in records:
        r.sequence = r.sequence.upper()
        if len(r.sequence) == 0:
            logger.info("Record %r is empty, skipping it.", r.name)
            continue
        name = r.name.split()[0]
        i = 0
        while name in names:
            i += 1
            name = r.name + '_{}'.format(i)
        if name != r.name.split()[0]:
            logger.info('Record name %r replaced with %r because it occurs more than once', r.name, name)
        if r.sequence in sequences:
            logger.info('Skipping %r because it contains the same sequence as %r',
                r.name, sequences[r.sequence])
            continue
        sequences[r.sequence] = name
        names.add(name)
        r.name = name
        yield r


def main(args):
    if ' ' in args.directory:
        logger.error('The name of the analysis directory must not contain spaces')
        sys.exit(1)

    if os.path.exists(args.directory):
        logger.error('The target directory {!r} already exists.'.format(args.directory))
        sys.exit(1)

    # If reads files or database were not given, initialize the GUI
    if (args.reads1 is None and args.single_reads is None) or args.database is None:
        try:
            gui = TkinterGui()
        except ImportError:  # TODO tk.TclError cannot be caught when import of tk fails
            logger.error('GUI cannot be started. Please provide reads1 file '
                'and database directory on command line.')
            sys.exit(1)
    else:
        gui = None

    # Find out whether data is paired or single
    assert not (args.reads1 and args.single_reads)
    if args.reads1 is None and args.single_reads is None:
        paired = gui.yesno('Paired end or single-end reads',
            'Are your reads paired and need to be merged?\n\n'
            'If you answer "Yes", next select the FASTQ files '
            'with the first of your paired-end reads.\n'
            'If you answer "No", next select the FASTA or FASTQ '
            'file with single-end reads.')
        if paired is None:
            logger.error('Cancelled')
            sys.exit(2)
    else:
        paired = bool(args.reads1)

    # Assign reads1 and (if paired) also reads2
    if paired:
        if args.reads1 is not None:
            reads1 = args.reads1
            try_open(reads1)
        else:
            reads1 = gui.reads1_path()
            if not reads1:
                logger.error('Cancelled')
                sys.exit(2)
        reads2 = guess_paired_path(reads1)
        if reads2 is None:
            logger.error('Could not determine second file of paired-end reads')
            sys.exit(1)
    else:
        if args.single_reads is not None:
            reads1 = args.single_reads
            try_open(reads1)
        else:
            reads1 = gui.single_reads_path()
            if not reads1:
                logger.error('Cancelled')
                sys.exit(2)

    if args.database is not None:
        dbpath = args.database
    else:
        # TODO as soon as we distribute our own database files, we can use this:
        # database_path = pkg_resources.resource_filename('igdiscover', 'databases')
        databases_path = None
        dbpath = gui.database_path(databases_path)
        if not dbpath:
            logger.error('Cancelled')
            sys.exit(2)

    database = dict()
    for g in ['V', 'D', 'J']:
        path = os.path.join(dbpath, g + '.fasta')
        if not os.path.exists(path):
            logger.error(
                'The database directory %r must contain the three files '
                'V.fasta, D.fasta and J.fasta', dbpath)
            logger.error(
                'A dummy D.fasta is necessary even if analyzing light chains (see manual)')
            sys.exit(2)
        database[g] = list(read_and_repair_fasta(path))

    # Create the directory
    try:
        os.mkdir(args.directory)
    except OSError as e:
        logger.error(e)
        sys.exit(1)

    def create_symlink(readspath, dirname, target):
        gz = '.gz' if readspath.endswith('.gz') else ''
        if not os.path.isabs(readspath):
            src = os.path.relpath(readspath, dirname)
        else:
            src = readspath
        os.symlink(src, os.path.join(dirname, target + gz))

    if paired:
        create_symlink(reads1, args.directory, 'reads.1.fastq')
        create_symlink(reads2, args.directory, 'reads.2.fastq')
    else:
        try:
            target = 'reads.' + file_type(reads1)
        except UnknownFileFormatError:
            logger.error('Cannot determine whether reads file is FASTA or FASTQ')
            sys.exit(1)
        create_symlink(reads1, args.directory, target)

    # Write the configuration file
    configuration = pkg_resources.resource_string('igdiscover', Config.DEFAULT_PATH).decode()
    with open(os.path.join(args.directory, Config.DEFAULT_PATH), 'w') as f:
        f.write(configuration)

    # Create database files
    database_dir = os.path.join(args.directory, 'database')
    os.mkdir(database_dir)
    for gene in ['V', 'D', 'J']:
        with open(os.path.join(database_dir, gene + '.fasta'), 'w') as db_file:
            for record in database[gene]:
                print('>{}\n{}'.format(record.name, record.sequence), file=db_file)

    if gui is not None:
        # Only suggest to edit the config file if at least one GUI dialog has been shown
        if gui.yesno('Directory initialized',
                'Do you want to edit the configuration file now?'):
            launch(os.path.join(args.directory, Config.DEFAULT_PATH))
    logger.info('Directory %s initialized.', args.directory)
    logger.info('Edit %s/%s, then run "cd %s && igdiscover run" to start the analysis', args.directory, Config.DEFAULT_PATH, args.directory)



"""
Rename and reorder records in a FASTA file

Sequences can be renamed according to the sequences in a template file.
The sequences in the target file will get the name that they have in the
template file. Sequences are considered to be equivalent if one is a prefix of
the other.

Sequences can also be ordered by name, either alphabetically or by
the order given in a template file. For comparison, only the 'gene part'
of the name is used (before the '*'). For example, for 'VH4-4*01', the name 'VH4-4'
is looked up in the template. Alphabetic order is the default. Use --no-sort
to disable sorting entirely.
"""
import sys
import logging
import dnaio

from ..utils import natural_sort_key

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--no-sort', dest='sort', action='store_false', default=True,
        help='Do not sort sequences by name')
    arg('--not-found', metavar='TEXT', default=' (not found)',
        help='Append this text to the record name when the sequence was not found '
        'in the template. Default: %(default)r')
    arg('--rename-from', metavar='TEMPLATE',
        help='FASTA template file with correctly named sequences. If a sequence in '
            'the target file is identical to one in the template, it is assigned the '
            'name of the sequence in the template.')
    arg('--order-by', metavar='TEMPLATE',
        help='FASTA template that contains genes in the desired order. '
            'If a name contains a "*" (asterisk), only the preceding part '
            'is used. Thus, VH4-4*01 and VH4-4*02 are equivalent.')
    arg('target', help='FASTA file with to-be renamed sequences')


class PrefixDict:
    """
    A dict that maps strings to values, but where a prefix of a key is enough
    to retrieve the value.
    """
    def __init__(self, items):
        self._items = []
        for k, v in items:
            self.add(k, v)

    def add(self, k, v):
        try:
            self[k]
            raise ValueError('Key {!r} already exists'.format(k))
        except KeyError:
            self._items.append((k, v))

    def __getitem__(self, key):
        found = None
        for seq, value in self._items:
            if seq.startswith(key) or key.startswith(seq):
                if found is not None:
                    # TODO don't use keyerror here
                    raise KeyError('Key {!r} is ambiguous'.format(key))
                found = value
        if found is None:
            raise KeyError(key)
        return found

    def get(self, key, default=None):
        try:
            v = self[key]
        except KeyError:
            return default
        else:
            return v

    def __len__(self):
        return len(self._items)


class GeneMissing(Exception):
    pass


def gene_name(record):
    record_name, _, record_comment = record.name.partition(' ')
    return record_name.partition('*')[0]


def sorted_by_gene(records, gene_order):
    d = {name: i for i, name in enumerate(gene_order)}

    def keyfunc(record):
        gene = gene_name(record)
        try:
            return d[gene]
        except KeyError:
            raise GeneMissing(gene)

    return sorted(records, key=keyfunc)


def main(args):
    if args.rename_from:
        with dnaio.open(args.rename_from) as fr:
            template = PrefixDict([])
            for record in fr:
                try:
                    template.add(record.sequence.upper(), record.name)
                except ValueError:
                    logger.error('Sequences in entry %r and %r are duplicate',
                        record.name, template[record.sequence.upper()])
        logger.info('Read %d entries from template', len(template))
    else:
        template = None

    if args.order_by:
        with dnaio.open(args.order_by) as fr:
            gene_order = [gene_name(r) for r in fr]
    else:
        gene_order = None

    with dnaio.open(args.target) as fr:
        sequences = list(fr)

    # Rename
    renamed = 0
    if template is not None:
        for record in sequences:
            name = template.get(record.sequence.upper())
            if name is None:
                name = record.name + args.not_found
            else:
                renamed += 1
            # Replace record’s name, leaving comment intact
            record_name, _, record_comment = record.name.partition(' ')
            if record_comment:
                record.name = name + ' ' + record_comment
            else:
                record.name = name

    # Reorder
    if gene_order:
        try:
            sequences = sorted_by_gene(sequences, gene_order)
        except GeneMissing as e:
            logger.error('Gene "%s" not found in the --order-by template file', e)
            sys.exit(1)
    elif args.sort:
        sequences = sorted(sequences, key=lambda s: natural_sort_key(s.name))
    for record in sequences:
        print('>{}\n{}'.format(record.name, record.sequence))
    logger.info('Wrote %s FASTA records (%d sequences found in template)', len(sequences), renamed)
